import json
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig import instrument_events as ctt
from fbnconfig import property as prop

TEST_BASE = "https://foo.lusid.com"


def response_hook(response: httpx.Response):
    response.read()
    response.raise_for_status()


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeTransactionTemplate:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_create(self, respx_mock):

        scope = "test_scope"
        instrument_event_type = "BondCouponEvent"
        instrument_type = "Bond"

        respx_mock.post(
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{scope}"
        ).mock(return_value=httpx.Response(200, json={}))

        sut = ctt.TransactionTemplateResource(
            id="template_id",
            scope=scope,
            instrument_type=instrument_type,
            instrument_event_type=instrument_event_type,
            description="A test transaction template",
            component_transactions=[
                ctt.ComponentTransactions(
                    display_name="Bond Income Override",
                    condition="{{eligibleBalance}} gt 200",
                    transaction_field_map=ctt.TransactionFieldMap(
                        instrument="{{instrument}}",
                        settlement_date="{{BondCouponEvent.paymentDate}}",
                        source="MyTransactionTypeSource",
                        transaction_currency="{{BondCouponEvent.currency}}",
                        transaction_date="{{BondCouponEvent.exDate}}",
                        transaction_id=(
                            "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
                        ),
                        type="BondCoupon",
                        units="{{eligibleBalance}}",
                        transaction_price=ctt.TransactionPriceAndType(
                            price="{{BondCouponEvent.couponPerUnit}}",
                            type="CashFlowPerUnit"
                        ),
                        exchange_rate="1",
                        total_consideration=ctt.TransactionCurrencyAndAmount(
                            currency="{{BondCouponEvent.currency}}",
                            amount="{{BondCouponEvent.couponAmount}}"
                        ),
                        ),
                    transaction_property_map=[],
                    preserve_tax_lot_structure=None,
                    market_open_time_adjustments=None
                )
        ])

        state = sut.create(self.client)
        assert state["id"] == "template_id"
        assert state["scope"] == scope
        assert state["instrumentType"] == instrument_type
        assert state["instrumentEventType"] == instrument_event_type

        request = respx_mock.calls.last.request

        assert request.method == "POST"
        assert request.url.path == (
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{scope}"
        )
        request_body = json.loads(request.content)

        assert request_body["scope"] == scope
        assert request_body["instrumentType"] == instrument_type
        assert request_body["instrumentEventType"] == instrument_event_type
        assert request_body["description"] == "A test transaction template"

        # Assert componentTransactions structure
        assert "componentTransactions" in request_body
        assert len(request_body["componentTransactions"]) == 1

        component_txn = request_body["componentTransactions"][0]
        assert component_txn["displayName"] == "Bond Income Override"
        assert component_txn["condition"] == "{{eligibleBalance}} gt 200"

        # Assert transactionFieldMap
        expected_txn_field_map = {
            "instrument": "{{instrument}}",
            "settlementDate": "{{BondCouponEvent.paymentDate}}",
            "source": "MyTransactionTypeSource",
            "transactionCurrency": "{{BondCouponEvent.currency}}",
            "transactionDate": "{{BondCouponEvent.exDate}}",
            "transactionId": "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}",
            "type": "BondCoupon",
            "units": "{{eligibleBalance}}",
            "exchangeRate": "1",
            "transactionPrice": {
                "price": "{{BondCouponEvent.couponPerUnit}}",
                "type": "CashFlowPerUnit"
            },
            "totalConsideration": {
                "currency": "{{BondCouponEvent.currency}}",
                "amount": "{{BondCouponEvent.couponAmount}}"
            }
        }
        assert component_txn["transactionFieldMap"] == expected_txn_field_map

        # Assert other fields
        assert component_txn["transactionPropertyMap"] == []
        assert (
            "preserveTaxLotStructure" not in component_txn or
            component_txn["preserveTaxLotStructure"] is None
        )
        assert (
            "marketOpenTimeAdjustments" not in component_txn or
            component_txn["marketOpenTimeAdjustments"] is None
        )

    def test_create_with_props(self, respx_mock):

        scope = "test_scope_with_props"
        instrument_event_type = "BondCouponEvent"
        instrument_type = "Bond"

        respx_mock.post(
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{scope}"
        ).mock(return_value=httpx.Response(200, json={}))

        sut = ctt.TransactionTemplateResource(
            id="template_id_with_props",
            scope=scope,
            instrument_type=instrument_type,
            instrument_event_type=instrument_event_type,
            description="A test transaction template with properties",
            component_transactions=[
                ctt.ComponentTransactions(
                    display_name="Bond Income Override",
                    condition="{{eligibleBalance}} gt 200",
                    transaction_field_map=ctt.TransactionFieldMap(
                        instrument="{{instrument}}",
                        settlement_date="{{BondCouponEvent.paymentDate}}",
                        source="MyTransactionTypeSource",
                        transaction_currency="{{BondCouponEvent.currency}}",
                        transaction_date="{{BondCouponEvent.exDate}}",
                        transaction_id=(
                            "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
                        ),
                        type="BondCoupon",
                        units="{{eligibleBalance}}",
                        transaction_price=ctt.TransactionPriceAndType(
                            price="{{BondCouponEvent.couponPerUnit}}",
                            type="CashFlowPerUnit"
                        ),
                        exchange_rate="1",
                        total_consideration=ctt.TransactionCurrencyAndAmount(
                            currency="{{BondCouponEvent.currency}}",
                            amount="{{BondCouponEvent.couponAmount}}"
                        ),
                        ),
                    transaction_property_map=[
                        ctt.TransactionPropertyMap(
                            property_key=prop.DefinitionRef(
                                id="currency_property_ref",
                                domain=prop.Domain.Transaction,
                                scope="test-scope",
                                code="MyCurrencyProperty"
                            ),
                            value="{{BondCouponEvent.currency}}"
                        ),
                        ctt.TransactionPropertyMap(
                            property_key=prop.DefinitionRef(
                                id="instrument_property_ref",
                                domain=prop.Domain.Transaction,
                                scope="test-scope",
                                code="MyInstrumentProperty"
                            ),
                            value="{{instrument}}"
                        )
                    ],
                    preserve_tax_lot_structure=None,
                    market_open_time_adjustments=None
                )
        ])

        state = sut.create(self.client)
        assert state["scope"] == scope
        assert state["instrumentType"] == instrument_type
        assert state["instrumentEventType"] == instrument_event_type

        request = respx_mock.calls.last.request

        assert request.method == "POST"
        assert request.url.path == (
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{scope}"
        )
        request_body = json.loads(request.content)

        assert request_body["scope"] == scope
        assert request_body["instrumentType"] == instrument_type
        assert request_body["instrumentEventType"] == instrument_event_type
        assert request_body["description"] == "A test transaction template with properties"

        # Assert componentTransactions structure
        assert "componentTransactions" in request_body
        assert len(request_body["componentTransactions"]) == 1

        component_txn = request_body["componentTransactions"][0]
        assert component_txn["displayName"] == "Bond Income Override"
        assert component_txn["condition"] == "{{eligibleBalance}} gt 200"

        # Assert transactionFieldMap
        expected_txn_field_map = {
            "instrument": "{{instrument}}",
            "settlementDate": "{{BondCouponEvent.paymentDate}}",
            "source": "MyTransactionTypeSource",
            "transactionCurrency": "{{BondCouponEvent.currency}}",
            "transactionDate": "{{BondCouponEvent.exDate}}",
            "transactionId": "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}",
            "type": "BondCoupon",
            "units": "{{eligibleBalance}}",
            "exchangeRate": "1",
            "transactionPrice": {
                "price": "{{BondCouponEvent.couponPerUnit}}",
                "type": "CashFlowPerUnit"
            },
            "totalConsideration": {
                "currency": "{{BondCouponEvent.currency}}",
                "amount": "{{BondCouponEvent.couponAmount}}"
            }
        }
        assert component_txn["transactionFieldMap"] == expected_txn_field_map

        # Assert transactionPropertyMap
        txn_property_map = component_txn["transactionPropertyMap"]
        assert len(txn_property_map) == 2
        assert txn_property_map[0]["propertyKey"] == "Transaction/test-scope/MyCurrencyProperty"
        assert txn_property_map[0]["value"] == "{{BondCouponEvent.currency}}"
        assert txn_property_map[1]["propertyKey"] == "Transaction/test-scope/MyInstrumentProperty"
        assert txn_property_map[1]["value"] == "{{instrument}}"

        # Assert other fields
        assert (
            "preserveTaxLotStructure" not in component_txn or
            component_txn["preserveTaxLotStructure"] is None
        )
        assert (
            "marketOpenTimeAdjustments" not in component_txn or
            component_txn["marketOpenTimeAdjustments"] is None
        )

    def test_update_with_no_changes(self, respx_mock):
        scope = "test_scope"
        instrument_event_type = "BondCouponEvent"
        instrument_type = "Bond"

        sut = ctt.TransactionTemplateResource(
            id="template_id",
            scope=scope,
            instrument_type=instrument_type,
            instrument_event_type=instrument_event_type,
            description="A test transaction template",
            component_transactions=[
                ctt.ComponentTransactions(
                    display_name="Bond Income Override",
                    condition="{{eligibleBalance}} gt 200",
                    transaction_field_map=ctt.TransactionFieldMap(
                        instrument="{{instrument}}",
                        settlement_date="{{BondCouponEvent.paymentDate}}",
                        source="MyTransactionTypeSource",
                        transaction_currency="{{BondCouponEvent.currency}}",
                        transaction_date="{{BondCouponEvent.exDate}}",
                        transaction_id=(
                            "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
                        ),
                        type="BondCoupon",
                        units="{{eligibleBalance}}",
                        transaction_price=ctt.TransactionPriceAndType(
                            price="{{BondCouponEvent.couponPerUnit}}",
                            type="CashFlowPerUnit"
                        ),
                        exchange_rate="1",
                        total_consideration=ctt.TransactionCurrencyAndAmount(
                            currency="{{BondCouponEvent.currency}}",
                            amount="{{BondCouponEvent.couponAmount}}"
                        ),
                        ),
                    transaction_property_map=[],
                    preserve_tax_lot_structure=None,
                    market_open_time_adjustments=None
                )
        ])

        desired = sut.model_dump(mode="json", exclude_none=True, by_alias=True)

        sorted_desired = json.dumps(desired, sort_keys=True)
        from hashlib import sha256
        content_hash = sha256(sorted_desired.encode()).hexdigest()

        old_state = SimpleNamespace(
            scope=scope,
            instrumentEventType=instrument_event_type,
            instrumentType=instrument_type,
            content_hash=content_hash
        )

        state = sut.update(self.client, old_state)
        assert state is None
        assert len(respx_mock.calls) == 0

    def test_update_with_changes(self, respx_mock):
        scope = "test_scope"
        instrument_event_type = "BondCouponEvent"
        instrument_type = "Bond"

        respx_mock.delete(
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{scope}"
        ).mock(return_value=httpx.Response(200, json={}))
        respx_mock.post(
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{scope}"
        ).mock(return_value=httpx.Response(200, json={}))

        sut = ctt.TransactionTemplateResource(
            id="template_id",
            scope=scope,
            instrument_type=instrument_type,
            instrument_event_type=instrument_event_type,
            description="A test transaction template",
            component_transactions=[
                ctt.ComponentTransactions(
                    display_name="Bond Income Override",
                    condition="{{eligibleBalance}} gt 200",
                    transaction_field_map=ctt.TransactionFieldMap(
                        instrument="{{instrument}}",
                        settlement_date="{{BondCouponEvent.paymentDate}}",
                        source="MyTransactionTypeSource",
                        transaction_currency="{{BondCouponEvent.currency}}",
                        transaction_date="{{BondCouponEvent.exDate}}",
                        transaction_id=(
                            "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
                        ),
                        type="BondCoupon",
                        units="{{eligibleBalance}}",
                        transaction_price=ctt.TransactionPriceAndType(
                            price="{{BondCouponEvent.couponPerUnit}}",
                            type="CashFlowPerUnit"
                        ),
                        exchange_rate="1",
                        total_consideration=ctt.TransactionCurrencyAndAmount(
                            currency="{{BondCouponEvent.currency}}",
                            amount="{{BondCouponEvent.couponAmount}}"
                        ),
                        ),
                    transaction_property_map=[],
                    preserve_tax_lot_structure=None,
                    market_open_time_adjustments=None
                )
        ])

        old_state = SimpleNamespace(
            scope=scope,
            instrumentEventType=instrument_event_type,
            instrumentType=instrument_type,
            content_hash="different_hash"
        )
        state = sut.update(self.client, old_state)

        assert state is not None
        assert len(respx_mock.calls) == 2
        assert state["scope"] == scope
        assert state["instrumentEventType"] == instrument_event_type
        assert state["instrumentType"] == instrument_type
        assert "content_hash" in state
        assert state["content_hash"] != "different_hash"

        # Verify the delete request
        delete_request = respx_mock.calls[0].request
        assert delete_request.method == "DELETE"
        assert delete_request.url.path == (
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{scope}"
        )

        post_request = respx_mock.calls[1].request
        assert post_request.method == "POST"
        assert post_request.url.path == (
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{scope}"
        )

    def test_update_with_scope_change(self, respx_mock):
        old_scope = "old_test_scope"
        new_scope = "new_test_scope"
        instrument_event_type = "BondCouponEvent"
        instrument_type = "Bond"

        respx_mock.delete(
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{old_scope}"
        ).mock(return_value=httpx.Response(200, json={}))
        respx_mock.post(
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{new_scope}"
        ).mock(return_value=httpx.Response(200, json={}))

        # Create a transaction template with new scope
        sut = ctt.TransactionTemplateResource(
            id="template_id",
            scope=new_scope,  # Different scope than old state
            instrument_type=instrument_type,
            instrument_event_type=instrument_event_type,
            description="A test transaction template",
            component_transactions=[
                ctt.ComponentTransactions(
                    display_name="Bond Income Override",
                    condition="{{eligibleBalance}} gt 200",
                    transaction_field_map=ctt.TransactionFieldMap(
                        instrument="{{instrument}}",
                        settlement_date="{{BondCouponEvent.paymentDate}}",
                        source="MyTransactionTypeSource",
                        transaction_currency="{{BondCouponEvent.currency}}",
                        transaction_date="{{BondCouponEvent.exDate}}",
                        transaction_id=(
                            "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
                        ),
                        type="BondCoupon",
                        units="{{eligibleBalance}}",
                        transaction_price=ctt.TransactionPriceAndType(
                            price="{{BondCouponEvent.couponPerUnit}}",
                            type="CashFlowPerUnit"
                        ),
                        exchange_rate="1",
                        total_consideration=ctt.TransactionCurrencyAndAmount(
                            currency="{{BondCouponEvent.currency}}",
                            amount="{{BondCouponEvent.couponAmount}}"
                        ),
                        ),
                    transaction_property_map=[],
                    preserve_tax_lot_structure=None,
                    market_open_time_adjustments=None
                )
        ])

        # Create old state with different scope
        old_state = SimpleNamespace(
            scope=old_scope,
            instrumentEventType=instrument_event_type,
            instrumentType=instrument_type,
            content_hash="old_hash"
        )

        # When we update it
        state = sut.update(self.client, old_state)

        # Then the old one is deleted and a new one is created
        assert len(respx_mock.calls) == 2
        delete_request = respx_mock.calls[0].request
        assert delete_request.method == "DELETE"
        assert delete_request.url.path == (
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{old_scope}"
        )

        post_request = respx_mock.calls[1].request
        assert post_request.method == "POST"
        assert post_request.url.path == (
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{new_scope}"
        )
        assert state is not None
        assert state["scope"] == new_scope
        assert state["instrumentEventType"] == instrument_event_type
        assert state["instrumentType"] == instrument_type
        assert "content_hash" in state

    def test_delete(self, respx_mock):
        scope = "test_scope"
        instrument_event_type = "BondCouponEvent"
        instrument_type = "Bond"

        # Mock the DELETE request
        respx_mock.delete(
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{scope}"
        ).mock(return_value=httpx.Response(200, json={}))

        # Create old state to delete
        old_state = SimpleNamespace(
            scope=scope,
            instrumentEventType=instrument_event_type,
            instrumentType=instrument_type
        )

        # When we delete it
        ctt.TransactionTemplateResource.delete(self.client, old_state)

        # Then a DELETE request was made
        assert len(respx_mock.calls) == 1
        delete_request = respx_mock.calls[0].request
        assert delete_request.method == "DELETE"
        assert delete_request.url.path == (
            f"/api/api/instrumenteventtypes/{instrument_event_type}/"
            f"transactiontemplates/{instrument_type}/{scope}"
        )

    def test_deps_no_props(self):
        sut = ctt.TransactionTemplateResource(
            id="template_id",
            scope="test_scope",
            instrument_type="Bond",
            instrument_event_type="BondCouponEvent",
            description="A test transaction template",
            component_transactions=[
                ctt.ComponentTransactions(
                    display_name="Bond Income Override",
                    condition="{{eligibleBalance}} gt 200",
                    transaction_field_map=ctt.TransactionFieldMap(
                        instrument="{{instrument}}",
                        settlement_date="{{BondCouponEvent.paymentDate}}",
                        source="MyTransactionTypeSource",
                        transaction_currency="{{BondCouponEvent.currency}}",
                        transaction_date="{{BondCouponEvent.exDate}}",
                        transaction_id=(
                            "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
                        ),
                        type="BondCoupon",
                        units="{{eligibleBalance}}",
                        transaction_price=ctt.TransactionPriceAndType(
                            price="{{BondCouponEvent.couponPerUnit}}",
                            type="CashFlowPerUnit"
                        ),
                        exchange_rate="1",
                        total_consideration=ctt.TransactionCurrencyAndAmount(
                            currency="{{BondCouponEvent.currency}}",
                            amount="{{BondCouponEvent.couponAmount}}"
                        ),
                        ),
                    transaction_property_map=[],
                    preserve_tax_lot_structure=None,
                    market_open_time_adjustments=None
                )
        ])
        assert sut.deps() == []

    def test_deps_with_props(self):
        currency_prop = prop.DefinitionRef(
            id="currency_property_ref",
            domain=prop.Domain.Transaction,
            scope="test-scope",
            code="MyCurrencyProperty"
        )
        instrument_prop = prop.DefinitionRef(
            id="instrument_property_ref",
            domain=prop.Domain.Transaction,
            scope="test-scope",
            code="MyInstrumentProperty"
        )

        sut = ctt.TransactionTemplateResource(
            id="template_id_with_props",
            scope="test_scope_with_props",
            instrument_type="Bond",
            instrument_event_type="BondCouponEvent",
            description="A test transaction template with properties",
            component_transactions=[
                ctt.ComponentTransactions(
                    display_name="Bond Income Override",
                    condition="{{eligibleBalance}} gt 200",
                    transaction_field_map=ctt.TransactionFieldMap(
                        instrument="{{instrument}}",
                        settlement_date="{{BondCouponEvent.paymentDate}}",
                        source="MyTransactionTypeSource",
                        transaction_currency="{{BondCouponEvent.currency}}",
                        transaction_date="{{BondCouponEvent.exDate}}",
                        transaction_id=(
                            "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
                        ),
                        type="BondCoupon",
                        units="{{eligibleBalance}}",
                        transaction_price=ctt.TransactionPriceAndType(
                            price="{{BondCouponEvent.couponPerUnit}}",
                            type="CashFlowPerUnit"
                        ),
                        exchange_rate="1",
                        total_consideration=ctt.TransactionCurrencyAndAmount(
                            currency="{{BondCouponEvent.currency}}",
                            amount="{{BondCouponEvent.couponAmount}}"
                        ),
                        ),
                    transaction_property_map=[
                        ctt.TransactionPropertyMap(
                            property_key=currency_prop,
                            value="{{BondCouponEvent.currency}}"
                        ),
                        ctt.TransactionPropertyMap(
                            property_key=instrument_prop,
                            value="{{somevalue}}"
                        )
                    ],
                    preserve_tax_lot_structure=None,
                    market_open_time_adjustments=None
                )
        ])

        deps = sut.deps()
        assert len(deps) == 2
        assert currency_prop in deps
        assert instrument_prop in deps

    def test_dump(self):
        currency_prop = prop.DefinitionRef(
            id="currency_property_ref",
            domain=prop.Domain.Transaction,
            scope="test-scope",
            code="MyCurrencyProperty"
        )
        instrument_prop = prop.DefinitionRef(
            id="instrument_property_ref",
            domain=prop.Domain.Transaction,
            scope="test-scope",
            code="MyInstrumentProperty"
        )
        sut = ctt.TransactionTemplateResource(
            id="template_id_with_props",
            scope="test_scope_with_props",
            instrument_type="Bond",
            instrument_event_type="BondCouponEvent",
            description="A test transaction template with properties",
            component_transactions=[
                ctt.ComponentTransactions(
                    display_name="Bond Income Override",
                    condition="{{eligibleBalance}} gt 200",
                    transaction_field_map=ctt.TransactionFieldMap(
                        instrument="{{instrument}}",
                        settlement_date="{{BondCouponEvent.paymentDate}}",
                        source="MyTransactionTypeSource",
                        transaction_currency="{{BondCouponEvent.currency}}",
                        transaction_date="{{BondCouponEvent.exDate}}",
                        transaction_id=(
                            "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}"
                        ),
                        type="BondCoupon",
                        units="{{eligibleBalance}}",
                        transaction_price=ctt.TransactionPriceAndType(
                            price="{{BondCouponEvent.couponPerUnit}}",
                            type="CashFlowPerUnit"
                        ),
                        exchange_rate="1",
                        total_consideration=ctt.TransactionCurrencyAndAmount(
                            currency="{{BondCouponEvent.currency}}",
                            amount="{{BondCouponEvent.couponAmount}}"
                        ),
                        ),
                    transaction_property_map=[
                        ctt.TransactionPropertyMap(
                            property_key=currency_prop,
                            value="{{BondCouponEvent.currency}}"
                        ),
                        ctt.TransactionPropertyMap(
                            property_key=instrument_prop,
                            value="{{somevalue}}"
                        )
                    ],
                    preserve_tax_lot_structure=None,
                    market_open_time_adjustments=None
                )
        ])
        # when we dump it
        dump = sut.model_dump(by_alias=True, exclude_none=True,
                              round_trip=True, context={"style": "dump"})
        # the id is exluded
        assert dump.get("id", None) is None
        # The properties are serialized as $refs
        tpm = dump["componentTransactions"][0]["transactionPropertyMap"]
        assert tpm[0]["propertyKey"] == {"$ref": currency_prop.id}

    def test_undump(self):
        # given we are parsing a dump and we have already parsed the
        # properties
        currency_prop = prop.DefinitionRef(
            id="currency_property_ref",
            domain=prop.Domain.Transaction,
            scope="test-scope",
            code="MyCurrencyProperty"
        )
        instrument_prop = prop.DefinitionRef(
            id="instrument_property_ref",
            domain=prop.Domain.Transaction,
            scope="test-scope",
            code="MyInstrumentProperty"
        )
        # but we have to parse the transaction template
        dump = {
            "componentTransactions": [
              {
                "condition": "{{eligibleBalance}} gt 200",
                "displayName": "Bond Income Override",
                "transactionFieldMap": {
                  "exchangeRate": "1",
                  "instrument": "{{instrument}}",
                  "settlementDate": "{{BondCouponEvent.paymentDate}}",
                  "source": "MyTransactionTypeSource",
                  "totalConsideration": {
                    "amount": "{{BondCouponEvent.couponAmount}}",
                    "currency": "{{BondCouponEvent.currency}}"
                  },
                  "transactionCurrency": "{{BondCouponEvent.currency}}",
                  "transactionDate": "{{BondCouponEvent.exDate}}",
                  "transactionId": "Automatically-generated txn: {{instrumentEventId}}-{{holdingId}}",
                  "transactionPrice": {
                    "price": "{{BondCouponEvent.couponPerUnit}}",
                    "type": "CashFlowPerUnit"
                  },
                  "type": "BondCoupon",
                  "units": "{{eligibleBalance}}"
                },
                "transactionPropertyMap": [
                  {
                    "propertyKey": {
                      "$ref": "currency_property_ref"
                    },
                    "value": "{{BondCouponEvent.currency}}"
                  },
                  {
                    "propertyKey": {
                      "$ref": "instrument_property_ref"
                    },
                    "value": "{{somevalue}}"
                  }
                ]
              }
            ],
            "description": "A test transaction template with properties",
            "instrumentEventType": "BondCouponEvent",
            "instrumentType": "Bond",
            "scope": "test_scope_with_props"
          }
        # when we undump it
        sut = ctt.TransactionTemplateResource.model_validate(
            dump,
            context={
                "style": "dump",
                "id": "template-id",
                "$refs": {p.id: p for p in [currency_prop, instrument_prop]}
            }
        )
        # then the properties get resolved from the $refs
        tpm = sut.component_transactions[0].transaction_property_map
        assert tpm[0].property_key == currency_prop
        # and the resource id is extracted from the context
        assert sut.id == "template-id"



from fbnconfig import Deployment, scheduler


def configure(env) -> Deployment:
    deployment_name = getattr(env, "name", "schedule")
    tag = "v5.0"
    i = scheduler.ImageResource(
        id="img1",
        source_image="harbor.finbourne.com/ceng/fbnconfig-pipeline:0.1",
        dest_name="beany",
        dest_tag=tag,
    )
    j = scheduler.JobResource(
        id="job1",
        scope=deployment_name,
        code="job",
        image=i,
        name="RobsJob",
        description="something nice",
        min_cpu="1",
        max_cpu="2",
        argument_definitions={
            "arg1": scheduler.EnvironmentArg(
                data_type="String",
                required=False,
                description="My argument",
                order=1,
                default_value="mydefaultvalue",
            )
        },
    )
    s = scheduler.ScheduleResource(
        id="sch1",
        name="my-schedule",
        scope=deployment_name,
        code="python-sched",
        expression="0 30 4 * * ? *",
        timezone="Europe/London",
        job=j,
        description="my awesome schedule",
        arguments={"arg1": "mynotdefaultvalue"},
    )
    return Deployment(deployment_name, [i, j, s])


if __name__ == "__main__":
    import os

    import click

    import fbnconfig

    @click.command()
    @click.argument("lusid_url", envvar="LUSID_ENV", type=str)
    @click.option("-v", "--vars_file", type=click.File("r"))
    def cli(lusid_url, vars_file):
        host_vars = fbnconfig.load_vars(vars_file)
        d = configure(host_vars)
        fbnconfig.deploy(d, lusid_url, os.environ["FBN_ACCESS_TOKEN"])

    cli()



import datetime

from fbnconfig.urlfmt import Urlfmt


def test_base():
    # given a base with a slash in it
    fmt = Urlfmt("http://www.foo.com/path")
    # when it is expanded
    res = fmt.format("{base}/{p1}/p2", **{"p1": "some thing"})
    # then the base path does not get encoded and the slash is literal
    assert res == "http://www.foo.com/path/some%20thing/p2"


def test_dict():
    fmt = Urlfmt("http://www.google.com")
    res = fmt.format("{base}/{p1}/p2", **{"p1": "some thing"})
    assert res == "http://www.google.com/some%20thing/p2"


def test_keywords():
    fmt = Urlfmt("http://www.google.com")
    scope = "scope1"
    code = "a code"
    res = fmt.format("{scope}/{code}", scope=scope, code=code)
    assert res == "scope1/a%20code"


def test_slash():
    fmt = Urlfmt("http://www.google.com")
    res = fmt.format("/{p1}/p2", **{"p1": "some/thing"})
    assert res == "/some%2Fthing/p2"


def test_formatting():
    fmt = Urlfmt("http://www.google.com")
    # given a format which adds a leading +
    # when we format
    res = fmt.format("{number:+d}", **{"number": 100})
    # then the + is added and also encoded
    assert res == "%2B100"


def test_date():
    fmt = Urlfmt("http://www.google.com")
    # given a date
    # when we format
    d = datetime.datetime(2010, 7, 4, 12, 15, 58)
    res = fmt.format("{:%Y-%m-%d}", d)
    # then the date is formatted
    assert res == "2010-07-04"


def test_class():
    class Foo:
        u = Urlfmt("/api/system")

        def __init__(self):
            self.code = "foo"
            self.scope = "banjo"

        def read(self):
            return self.u.format("{base}/{scope}/{code}", scope=self.scope, code=self.code)

    f = Foo()
    res = f.read()
    assert res == "/api/system/banjo/foo"



import hashlib
import json
import subprocess
import time
from collections import defaultdict
from datetime import datetime
from typing import Annotated, Any, Dict, List, Literal, Optional

import httpx
from httpx import Client, HTTPStatusError
from pydantic import (
    BaseModel,
    BeforeValidator,
    Field,
    PlainSerializer,
    computed_field,
    field_serializer,
    field_validator,
    model_validator,
)

from . import configuration as cfg
from . import identity
from .resource_abc import CamelAlias, Ref, Resource, register_resource


@register_resource()
class ImageRef(BaseModel, Ref):
    """Reference an existing image
    Example
    -------
    >>> from fbnconfig.scheduler import ImageRef
    >>> image_ref = ImageRef(dest_name="myimage", dest_tag="0.1.1")
    """

    id: str
    dest_name: str
    dest_tag: str

    def attach(self, client):
        # we only need to check it exists
        try:
            res = client.get(f"/scheduler2/api/images/repository/{self.dest_name}").json()
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Image with name {self.dest_name} and tag {self.dest_tag} not found")
            else:
                raise ex
        for img in res["values"]:
            names = {tag["name"] for tag in img["tags"]}
            if self.dest_tag in names:
                return
        raise RuntimeError(f"Image with name {self.dest_name} and tag {self.dest_tag} not found")


@register_resource()
class ImageResource(BaseModel, Resource):
    """Define an image

    Example
    -------
    >>> from fbnconfig.scheduler import ImageResource
    >>> image = ImageResource(
            id="img1",
            source_image="docker.io/alpine:3.16.7",
            dest_name="myimage",
            dest_tag="3.16.7")
    """

    id: str
    source_image: str
    dest_name: str
    dest_tag: str
    pull_upstream: bool | None = Field(default=True, exclude=True)

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    def read(self, client, old_state):
        pass  # not needed

    def create(self, client):
        # create a unique tag so scheduler will give us commands and not complain about
        # reusing an existing one. We want to check hashes not tags
        body = {"imageName": f"{self.dest_name}:{time.time()}"}
        upload = client.post("/scheduler2/api/images", json=body).json()
        auth_cmd = upload["dockerLoginCommand"].split(" ")
        user = auth_cmd[3]
        password = auth_cmd[5]
        reghost = auth_cmd[6]
        tag_cmd = upload["tagVersionedDockerImageCommand"].split(" ")
        # combine the repo from the api with the real tag we want
        downstream_repo = tag_cmd[3].split(":")[0]
        downstream_image = ":".join([downstream_repo, self.dest_tag])
        # pull from the upstream and apply the downstream tag
        pull_cmd = self._pull_commands(self.source_image) if self.pull_upstream else []
        tag_cmd = self._tag_commands(downstream_image)
        for cmd in pull_cmd + tag_cmd:
            subprocess.run(cmd).check_returncode()
        # get local and downstream digests to see if they match
        local_digest = self._get_local_digest(downstream_image)
        # check if the local hash matches the remote hash. If so skip the rest
        if local_digest is not None:
            downstream_images = self.get_remote_tags(client, self.dest_name, local_digest)
            matching_tag = next((tag for tag in downstream_images if tag["name"] == self.dest_tag), None)
            if matching_tag is not None:  # this image/tag combo already exists
                print("image already exists in the remote with the same digest")
                return {
                    "id": self.id,
                    "source_image": self.source_image,
                    "dest_name": self.dest_name,
                    "dest_tag": self.dest_tag,
                }
        # push the tag we actually want to have using the auth from scheduler
        push_commands = self._push_commands(downstream_image, user, password, reghost)
        for cmd in push_commands:
            subprocess.run(cmd).check_returncode()
        return {
            "id": self.id,
            "source_image": self.source_image,
            "dest_name": self.dest_name,
            "dest_tag": self.dest_tag,
        }

    @staticmethod
    def get_remote_tags(client, dest_name, local_digest):
        repo = client.get(f"/scheduler2/api/images/repository/{dest_name}")
        # todo: this could be paged
        return next((img["tags"] for img in repo.json()["values"] if img["digest"] == local_digest), [])

    @staticmethod
    def _get_local_digest(tag):
        inspect = subprocess.run(
            ["docker", "inspect", "--format", "{{json .RepoDigests}}", tag],
            capture_output=True,
            text=True,
        )
        inspect.check_returncode()
        # inspect returns:
        #   [alpine@sha256:a8cb.. fbn-qa.lusid.com/fbn-qa/beany@sha256:b797...]
        repo = tag.split(":")[0]
        return next(
            (
                digest.split("@")[1]
                for digest in json.loads(inspect.stdout)
                if digest.split("@")[0] == repo
            ),
            None,
        )

    @staticmethod
    def _push_commands(downstream_image, user, password, reghost):
        return [
            ["docker", "login", "-u", user, "--password", password, reghost],
            ["docker", "push", "-q", downstream_image],
        ]

    def _pull_commands(self, source_image):
        return [["docker", "pull", "--platform", "linux/amd64", "-q", source_image]]

    def _tag_commands(self, tag):
        return [["docker", "tag", self.source_image, tag]]

    def update(self, client, old_state):
        if (
            [self.source_image, self.dest_name, self.dest_tag] ==
            [old_state.source_image, old_state.dest_name, old_state.dest_tag]
        ):
            return None
        else:
            return self.create(client)

    @staticmethod
    def delete(client, old_state):
        # no delete for images
        pass

    def deps(self):
        return []


def ser_image_key(value, info):
    if info.context and info.context.get("style") == "dump":
        return {"$ref": value.id}
    return value


def des_image_key(value, info):
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get("$refs"):
        ref = info.context["$refs"][value["$ref"]]
        return ref
    return value


ImageKey = Annotated[
    ImageResource | ImageRef,
    BeforeValidator(des_image_key), PlainSerializer(ser_image_key)
]


class CommandlineArg(BaseModel, CamelAlias):
    """Define a commandline ArgumentDefinition for a JobResource"""

    data_type: Literal["String", "Int"]
    required: Optional[bool] = False
    description: str
    order: int
    passed_as: Literal["CommandLine"] = "CommandLine"
    default_value: None | str = None


class EnvironmentArg(BaseModel, CamelAlias):
    """Define a environment ArgumentDefinition for a JobResource"""

    data_type: Literal["String", "Int", "SecureString", "Configuration"]
    required: Optional[bool] = False
    description: str
    order: int
    passed_as: Literal["EnvironmentVariable"] = "EnvironmentVariable"
    default_value: None | str | cfg.ItemRef | cfg.ItemResource = None

    @field_serializer("default_value", when_used="always")
    def ser_value(self, value, info) -> Dict[str, str] | str | None:
        style = info.context.get("style", "api") if info.context else "api"
        if style == "dump":
            if isinstance(value, (cfg.ItemRef, cfg.ItemResource)):
                return {"$ref": value.id}
            return value
        # api style print is as config:abc
        if value is None:
            return None
        if isinstance(value, (cfg.ItemRef, cfg.ItemResource)):
            return value.ref
        return value

    @field_validator("default_value", mode="before")
    @classmethod
    def des_value(cls, value, info):
        if info.context and info.context.get("$refs"):
            if isinstance(value, dict) and value.get("$ref"):
                return info.context["$refs"][value["$ref"]]
        return value


JobArgument = Annotated[CommandlineArg | EnvironmentArg, Field(discriminator="passed_as")]


@register_resource()
class JobRef(BaseModel, Ref):
    """Reference an existing scheduler job"""

    id: str = Field(exclude=True)
    scope: str
    code: str

    _content_hash: str | None = None

    def attach(self, client):
        # just check it exists
        search = client.get(
            "/scheduler2/api/jobs",
            params={"filter": f"jobId.scope eq '{self.scope}' and jobId.code eq '{self.code}'"},
        )

        current = next((job for job in search.json()["values"]), None)
        if current is None:
            raise RuntimeError(
                f"Failed to attach JobRef to job with scope={self.scope} and code={self.code}"
            )

        self._content_hash = hashlib.sha256(json.dumps(current, sort_keys=True).encode()).hexdigest()


class JobState(BaseModel):
    id: str
    scope: str
    code: str


@register_resource()
class JobResource(BaseModel, Resource):
    """Manage a JobDefinition"""

    id: str = Field(exclude=True)
    scope: str = Field(exclude=False, init=True)
    code: str = Field(exclude=False, init=True)
    image: ImageKey
    name: str
    author: Optional[str] = None
    date_created: Optional[datetime] = None
    description: str
    ttl: Optional[int] = None
    min_cpu: Optional[str] = None
    max_cpu: Optional[str] = None
    min_memory: Optional[str] = None
    max_memory: Optional[str] = None
    argument_definitions: Dict[str, JobArgument] = {}
    command_line_argument_separator: Optional[str] = None

    class ScanTimeoutParameters:
        tries = 120
        wait_time = 1

    @computed_field
    def job_id(self) -> Dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    @computed_field
    def image_name(self) -> str:
        return self.image.dest_name

    @computed_field
    def image_tag(self) -> str:
        return self.image.dest_tag

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Dict[str, Any], info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        # handle the list api returning dockerImage instead of image
        if data.get("dockerImage", None):
            data = data | {"image": data["dockerImage"]}
            data.pop("dockerImage")
        # Handle API list response format (jobIdd.scope and jobIdd.code)
        if isinstance(data.get("jobId", None), dict):
            data = data | data["jobId"]
            data.pop("jobId")
        # Handle id from context (for dump/undump)
        if info.context and info.context.get("id"):
            data = data | {"id": info.context.get("id")}
        return data

    @field_validator("image", mode="before")
    @classmethod
    def des_image(cls, image, info) -> ImageResource | ImageRef:
        if info.context and info.context.get("$refs"):
            return info.context["$refs"][image["$ref"]]
        return image

    def read(self, client: Client, old_state):
        search = client.get(
            "/scheduler2/api/jobs",
            params={
                "filter": f"jobId.scope eq '{old_state.scope}' and jobId.code eq '{old_state.code}'"
            },
        )
        current = next(job for job in search.json()["values"])
        # normalise the current values
        current.pop("jobId")
        current.pop("requiredResources")
        current["imageName"] = current["dockerImage"].split(":")[
            0
        ]  # we get dockerImage but have to send separately :(
        current["imageTag"] = current["dockerImage"].split(":")[1]
        current.pop("dockerImage")
        return current

    def _wait_for_vulnerability_scan(self, client):
        tries = self.ScanTimeoutParameters.tries
        while True:
            values = client.get(f"/scheduler2/api/images/repository/{self.image_name}").json()["values"]

            filtered_images = [v for v in values if self.image_tag in [tag["name"] for tag in v["tags"]]]

            if any(v["scanStatus"] == "COMPLETE" for v in filtered_images):
                break
            if any(
                (
                    v["scanStatus"]
                    in (
                        "FAILED",
                        "UNSUPPORTED_IMAGE",
                        "SCAN_ELIGIBILITY_EXPIRED",
                        "FINDINGS_UNAVAILABLE",
                    )
                    for v in filtered_images
                )
            ):
                raise RuntimeError("Image vulnerability scan failed.")
            time.sleep(self.ScanTimeoutParameters.wait_time)
            tries -= 1
            if tries < 1:
                raise RuntimeError("Image vulnerability scan timed out.")

    def create(self, client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True,
                                  exclude={"scope", "code", "image"})
        self._wait_for_vulnerability_scan(client)
        client.post("/scheduler2/api/jobs", json=desired)
        new_state = JobState(id=self.id, scope=self.scope, code=self.code)
        return new_state.model_dump()

    def update(self, client: Client, old_state):
        if self.scope != old_state.scope or self.code != old_state.code:
            raise (RuntimeError("Cannot change identifier on job. Create a new one"))

        remote = self.read(client, old_state)
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True,
                                  exclude={"job_id", "scope", "code", "image"})
        effective = remote | desired
        remote_args = remote["argumentDefinitions"]
        effective_args: dict[str, dict] = defaultdict(None)
        for arg_key, arg_value in desired["argumentDefinitions"].items():
            v: dict[str, str] = remote_args.get(arg_key, {}) | arg_value
            # if desired state does not have a default value, remove it from effective
            if arg_value.get("defaultValue", None) is None:
                v.pop("defaultValue", None)
            effective_args[arg_key] = v

        effective["argumentDefinitions"] = effective_args
        if effective == remote:
            return None
        self._wait_for_vulnerability_scan(client)
        client.put(f"/scheduler2/api/jobs/{self.scope}/{self.code}", json=desired)
        new_state = JobState(id=self.id, scope=self.scope, code=self.code)
        return new_state.model_dump()

    @staticmethod
    def delete(client: Client, old_state):
        client.delete(f"/scheduler2/api/jobs/{old_state.scope}/{old_state.code}")

    def cleanup(self, client, old_state):
        if self.scope != old_state.scope or self.code != old_state.code:
            return self.delete(client, old_state)
        return None

    def deps(self):
        config_args: List[cfg.ItemRef | cfg.ItemResource] = [
            arg.default_value
            for arg in self.argument_definitions.values()
            if isinstance(arg.default_value, (cfg.ItemRef, cfg.ItemResource))
        ]
        return [self.image] + config_args


def ser_job_key(value, info):
    if info.context and info.context.get("style") == "dump":
        return {"$ref": value.id}
    return {"scope": value.scope, "code": value.code}


def des_job_key(value, info):
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get("$refs"):
        ref = info.context["$refs"][value["$ref"]]
        return ref
    return value


JobKey = Annotated[
    JobResource | JobRef,
    BeforeValidator(des_job_key), PlainSerializer(ser_job_key)
]


class ScheduleState(BaseModel):
    id: str
    scope: str
    code: str
    # Suppress to keep backwards comp with existing state
    argKeys: List[str]  # noqa: N815


@register_resource()
class ScheduleResource(BaseModel, Resource):
    """Manage a ScheduleDefinition"""

    id: str = Field(exclude=True)
    name: str
    scope: str
    code: str
    expression: str
    timezone: Optional[str] = None
    job: JobKey
    description: str
    author: Optional[str] = None
    owner: Optional[str] = None
    arguments: None | Dict[str, str | cfg.ItemKey] = None
    enabled: Optional[bool] = None
    use_as_auth: identity.UserKey | None = None

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        # Handle API list response format (jobIdd.scope and jobIdd.code)
        if isinstance(data.get("scheduleIdentifier", None), dict):
            data = data | data["scheduleIdentifier"]
            data.pop("scheduleIdentifier")
        # api list format uses jobId.scope and jobId.code, we want job
        if isinstance(data.get("jobId", None), dict):
            data = data | {"job": data["jobId"]}
            data.pop("jobId")
        # promote the trigger fields up to the top and change case in timeZone
        if isinstance(data.get("trigger", None), dict):
            time_trigger = data["trigger"]["timeTrigger"]
            data = data | {
                "expression": time_trigger["expression"],
                "timezone": time_trigger["timeZone"]
            }
            data.pop("trigger")
        # Handle id from context (for dump/undump)
        if info.context and info.context.get("id"):
            data = data | {"id": info.context.get("id")}
        return data

    @computed_field
    def job_id(self) -> Dict[str, str]:
        """Return the job identifier as a dictionary"""
        return {"scope": self.job.scope, "code": self.job.code}

    @computed_field
    def schedule_id(self) -> Dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    @computed_field
    def trigger(self) -> Dict[str, Dict[str, str]]:
        time_trigger = {"timeTrigger": {"expression": self.expression}}

        if self.timezone is not None:
            time_trigger["timeTrigger"]["timeZone"] = self.timezone

        return time_trigger

    def read(self, client, old_state):
        get = client.get(f"/scheduler2/api/schedules/{old_state.scope}/{old_state.code}")
        current = get.json()
        # note scheduleId for create scheduleIdentifier for read :(
        current.pop("scheduleIdentifier")
        return current

    def create(self, client):
        body = self.model_dump(mode="json", exclude_none=True, by_alias=True,
           exclude={"job", "scope", "code", "expression", "timezone"}
        )
        client.post("/scheduler2/api/schedules", json=body)
        arg_keys = list(self.arguments.keys()) if self.arguments else []
        return ScheduleState(id=self.id, scope=self.scope, code=self.code, argKeys=arg_keys).model_dump()

    def update(self, client, old_state):
        if self.scope != old_state.scope or self.code != old_state.code:
            from_id = f"{old_state.scope}/{old_state.code}"
            to_id = f"{self.scope}/{self.code}"
            raise (RuntimeError(f"Cannot change schedule identifier. From: '{from_id}', to: '{to_id}'"))

        remote = self.read(client, old_state)
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True,
          exclude={"schedule_id", "job", "scope", "code", "expression", "timezone"}
        )
        effective = remote | desired
        # On read the schedule will contain schedule args and args from the job.
        # Only update if the user specified args are different to the remote
        effective["arguments"] = remote.get("arguments", {}) | desired.get("arguments", {})
        arg_keys = list(self.arguments.keys()) if self.arguments else []
        if effective == remote and arg_keys == getattr(old_state, "argKeys", []):
            return None
        client.put(f"/scheduler2/api/schedules/{self.scope}/{self.code}", json=desired)
        return ScheduleState(id=self.id, scope=self.scope, code=self.code, argKeys=arg_keys).model_dump()

    @staticmethod
    def delete(client, old_state):
        try:
            client.delete(f"/scheduler2/api/schedules/{old_state.scope}/{old_state.code}")
        except HTTPStatusError as ex:
            content = ex.response.json()
            if (
                ex.response.status_code == 404
                and content.get("name", None) == "ValidationError"
                and content.get("title", None) == "Schedule could not be found"
            ):
                pass  # don't throw if schedule was already deleted
            else:
                raise ex

    def deps(self):
        config_args: List[cfg.ItemRef | cfg.ItemResource] = (
            [
                value
                for value in self.arguments.values()
                if isinstance(value, (cfg.ItemRef, cfg.ItemResource))
            ]
            if self.arguments
            else []
        )

        return [self.job] + config_args + ([self.use_as_auth] if self.use_as_auth is not None else [])


@register_resource()
class ScheduleRef(BaseModel, Ref):
    """Used to reference an existing schedule """

    id: str = Field(exclude=True)
    scope: str
    code: str

    def attach(self, client):
        scope, code = self.scope, self.code
        try:
            client.get(f"/scheduler2/api/schedules/{scope}/{code}")
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Schedule {scope}/{code} not found")
            else:
                raise ex



from fbnconfig import Deployment, datatype, fund_accounting, posting_module, property

"""
An example configuration for defining fund accounting related entities.
The script configures the following entities:
- Chart of accounts
- Accounts within a chart of account
- Posting module
"""


def configure(env):

    rating = property.DefinitionResource(
        id="rating",
        domain=property.Domain.ChartOfAccounts,
        scope="sc1",
        code="rating",
        display_name="Rating",
        data_type_id=datatype.DataTypeRef(id="default_str", scope="system", code="string"),
        constraint_style=property.ConstraintStyle.Collection,
        property_description="Example property representing a rating",
        life_time=property.LifeTime.Perpetual,
        collection_type=property.CollectionType.Array,
    )

    chart_property = fund_accounting.PropertyValue(
        property_key=rating,
        label_value="Example_label"
    )

    chart_of_account = fund_accounting.ChartOfAccountsResource(
        id="example_chart",
        scope="example_scope",
        code="example_code",
        display_name="example_display_name",
        description="example_description",
        properties=[chart_property]
    )

    instrument_property_definition = property.DefinitionResource(
        id="pd1",
        domain=property.Domain.Account,
        scope="sc1",
        code="pd1",
        display_name="Property definition example",
        data_type_id=property.ResourceId(scope="system", code="number"),
        constraint_style=property.ConstraintStyle.Property,
        property_description="Example property definition",
        life_time=property.LifeTime.Perpetual,
        collection_type=None,
    )

    account_property = fund_accounting.PropertyValue(
        property_key=instrument_property_definition,
        label_value="Goodbye"
    )

    account = fund_accounting.AccountResource(
        id="example_account",
        chart_of_accounts=chart_of_account,
        code="account_code",
        description="example_desc",
        type=fund_accounting.AccountType.ASSET,
        status=fund_accounting.AccountStatus.ACTIVE,
        control="Manual",
        properties=[account_property]
    )

    rule = posting_module.PostingModuleRule(
        rule_id="example_rule",
        general_ledger_account_code=account,
        rule_filter="SourceType eq 'LusidTransaction'"
    )

    posting_mod = posting_module.PostingModuleResource(
        id="posting_module_id",
        chart_of_accounts=chart_of_account,
        code="module_code",
        display_name="example_display_name",
        description="example_description",
        rules=[rule]
    )

    return Deployment("fund_accounting_example", [chart_of_account, account, posting_mod])



import json
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig import fund_configurations as fc
from fbnconfig import property as pd
from fbnconfig.http_client import response_hook

TEST_BASE = "https://foo.lusid.com"


@pytest.fixture
def property_def():
    """Fixture for creating a property definition resource"""
    property_def_scope = "MyPropertyScope"
    return pd.DefinitionResource(
        id="property_definition_for_fund_configuration_unit_test",
        code="MyFundConfigProperty",
        data_type_id=pd.ResourceId(scope="system", code="string"),
        display_name="Fund configuration display name",
        domain=pd.Domain.FundConfiguration,
        scope=property_def_scope,
        life_time=pd.LifeTime.TimeVariant
    )


@pytest.fixture
def fund_configuration_with_properties(property_def):
    """Fixture for creating a fund configuration resource with properties"""
    return fc.FundConfigurationResource(
        id="fundconfig1",
        scope="scope-fund-configuration",
        code="code-fund-configuration",
        display_name="fund configuration display name",
        description="fund configuration description",
        dealing_filters=[
            fc.ComponentFilter(
                filter_id="SUB",
                filter="account.code startswith '3001'"
            ),
            fc.ComponentFilter(
                filter_id="RED",
                filter="account.code startswith '3002'"
            )
        ],
        pnl_filters=[
            fc.ComponentFilter(
                filter_id="SUB",
                filter="account.code startswith '3001'"
            ),
            fc.ComponentFilter(
                filter_id="RED",
                filter="account.code startswith '3002'"
            )
        ],
        back_out_filters=[
            fc.ComponentFilter(
                filter_id="SUB",
                filter="account.code startswith '3001'"
            ),
            fc.ComponentFilter(
                filter_id="RED",
                filter="account.code startswith '3002'"
            )
        ],
        external_fee_filters=[
            fc.ExternalFeeComponentFilter(
                filter_id="ShareClassAPnLAdjustment",
                filter="account.code eq 'ShareClassAPnl'",
                applies_to=fc.AppliesToEnum.PnLBucket
            )
        ],
        properties=[
            fc.PropertyValue.create(
                key=property_def,
                label_value="My fund configuration property value",
                effective_from="0001-01-01T00:00:00.0000000+00:00",
                    effective_until="9999-12-31T23:59:59.9999999+00:00"
                )
        ]
    )


@pytest.fixture
def fund_configuration_without_properties():
    """Fixture for creating a fund configuration resource without properties"""
    return fc.FundConfigurationResource(
        id="fundconfig1",
        scope="scope-fund-configuration",
        code="code-fund-configuration",
        display_name="fund configuration display name",
        description="fund configuration description",
        dealing_filters=[
            fc.ComponentFilter(
                filter_id="SUB",
                filter="account.code startswith '3001'"
            ),
            fc.ComponentFilter(
                filter_id="RED",
                filter="account.code startswith '3002'"
            )
        ],
        pnl_filters=[
            fc.ComponentFilter(
                filter_id="SUB",
                filter="account.code startswith '3001'"
            ),
            fc.ComponentFilter(
                filter_id="RED",
                filter="account.code startswith '3002'"
            )
        ],
        back_out_filters=[
            fc.ComponentFilter(
                filter_id="SUB",
                filter="account.code startswith '3001'"
            ),
            fc.ComponentFilter(
                filter_id="RED",
                filter="account.code startswith '3002'"
            )
        ],
        external_fee_filters=[
            fc.ExternalFeeComponentFilter(
                filter_id="ShareClassAPnLAdjustment",
                filter="account.code eq 'ShareClassAPnl'",
                applies_to=fc.AppliesToEnum.PnLBucket
            )
        ]
    )


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeFundConfiguration:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_create(self, respx_mock, fund_configuration_with_properties, property_def):
        base_url = "api/api/fundconfigurations"
        scope = "scope-fund-configuration"
        respx_mock.post(f"{TEST_BASE}/{base_url}/{scope}").mock(
            return_value=httpx.Response(201, json={})
        )

        sut = fund_configuration_with_properties
        state = sut.create(self.client)

        expected_dealings_filters = [
            {
                "filterId": "SUB",
                "filter": "account.code startswith '3001'"
            },
            {
                "filterId": "RED",
                "filter": "account.code startswith '3002'"
            }
        ]

        expected_pnl_filters = [
            {
                "filterId": "SUB",
                "filter": "account.code startswith '3001'"
            },
            {
                "filterId": "RED",
                "filter": "account.code startswith '3002'"
            }
        ]
        expected_back_out_filters = [
            {"filterId": "SUB",
             "filter": "account.code startswith '3001'"},
            {"filterId": "RED",
             "filter": "account.code startswith '3002'"}
        ]
        expected_external_fee_filters = [
            {
                "filterId": "ShareClassAPnLAdjustment",
                "filter": "account.code eq 'ShareClassAPnl'",
                "appliesTo": fc.AppliesToEnum.PnLBucket.value
            }
        ]

        property_def_scope = "MyPropertyScope"
        expected_properties = {
            f"FundConfiguration/{property_def_scope}/MyFundConfigProperty": {
                "key": f"FundConfiguration/{property_def_scope}/MyFundConfigProperty",
                "value": {
                    "labelValue": "My fund configuration property value"
                },
                # The effective values are not present in the
                "effectiveFrom": "0001-01-01T00:00:00.0000000+00:00",
                "effectiveUntil": "9999-12-31T23:59:59.9999999+00:00"
            }
        }

        assert state["code"] == "code-fund-configuration"
        assert state["scope"] == "scope-fund-configuration"
        assert "content_hash" in state

        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == f"/api/api/fundconfigurations/{scope}"
        request_body = json.loads(request.content)

        assert request_body["displayName"] == "fund configuration display name"
        assert request_body["description"] == "fund configuration description"
        assert request_body["dealingFilters"] == expected_dealings_filters
        assert request_body["pnlFilters"] == expected_pnl_filters
        assert request_body["backOutFilters"] == expected_back_out_filters
        assert request_body["externalFeeFilters"] == expected_external_fee_filters
        assert request_body["properties"] == expected_properties

    def test_deps_with_properties(self, respx_mock, fund_configuration_with_properties, property_def):
        sut = fund_configuration_with_properties
        deps = sut.deps()

        assert len(deps) == 1
        assert deps[0] == property_def

    def test_deps_with_no_properties(self, respx_mock, fund_configuration_without_properties):
        sut = fund_configuration_without_properties
        deps = sut.deps()

        assert len(deps) == 0
        assert deps == []

    def test_update_with_changes(self, respx_mock, fund_configuration_with_properties):
        base_url = "api/api/fundconfigurations"
        scope = "scope-fund-configuration"
        code = "code-fund-configuration"
        old_scope = "replaced-scope"
        old_code = "replaced-code"
        respx_mock.post(f"{TEST_BASE}/{base_url}/{scope}").mock(
            return_value=httpx.Response(201, json={})
        )
        respx_mock.delete(f"{TEST_BASE}/{base_url}/{old_scope}/{old_code}").mock(
            return_value=httpx.Response(200, json={})
        )

        old_state = SimpleNamespace(scope=old_scope, code=old_code, content_hash="different_hash")

        sut = fund_configuration_with_properties

        updated_state = sut.update(self.client, old_state)

        delete_request = respx_mock.calls[0].request
        post_request = respx_mock.calls[1].request
        assert delete_request.method == "DELETE"
        assert delete_request.url.path == f"/api/api/fundconfigurations/{old_scope}/{old_code}"
        assert post_request.method == "POST"
        assert post_request.url.path == f"/api/api/fundconfigurations/{scope}"
        assert updated_state is not None
        assert updated_state["content_hash"] != "different_hash"
        assert updated_state["scope"] == scope
        assert updated_state["code"] == code

    def test_update_with_no_changes(self, respx_mock, fund_configuration_with_properties):
        sut = fund_configuration_with_properties
        scope = "scope-fund-configuration"
        code = "code-fund-configuration"

        desired_state = sut.model_dump(mode="json", exclude_none=True, by_alias=True)

        sorted_desired = json.dumps(desired_state, sort_keys=True)
        from hashlib import sha256
        content_hash = sha256(sorted_desired.encode()).hexdigest()
        old_state = SimpleNamespace(scope=scope, code=code, content_hash=content_hash)
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None (no changes needed)
        assert state is None
        # and no HTTP requests were made
        assert len(respx_mock.calls) == 0

    def test_delete(self, respx_mock):
        base_url = "api/api/fundconfigurations"
        scope = "scope-fund-configuration"
        code = "code-fund-configuration"
        respx_mock.delete(f"{TEST_BASE}/{base_url}/{scope}/{code}").mock(
            return_value=httpx.Response(201, json={})
        )

        old_state = SimpleNamespace(scope=scope, code=code, content_hash="some_hash")

        fc.FundConfigurationResource.delete(self.client, old_state)
        assert len(respx_mock.calls) == 1
        assert respx_mock.calls[0].request.method == "DELETE"
        assert respx_mock.calls[0].request.url.path == f"/api/api/fundconfigurations/{scope}/{code}"

    def test_dump_properties_with_fund_configuration(
        self, fund_configuration_with_properties: fc.FundConfigurationResource,
        property_def: pd.DefinitionResource
    ):
        sut = fund_configuration_with_properties
        dump = sut.model_dump(by_alias=True, exclude_none=True,
                              round_trip=True, context={"style": "dump"})

        property_key = list(dump["properties"].keys())[0]
        property_value = dump["properties"][property_key]
        assert "properties" in dump
        assert len(dump["properties"]) == 1
        assert property_key == "FundConfiguration/MyPropertyScope/MyFundConfigProperty"
        assert property_value["key"] == {
            "$ref": property_def.id
        }
        assert property_value["value"]["labelValue"] == "My fund configuration property value"
        assert property_value["effectiveFrom"] == "0001-01-01T00:00:00.0000000+00:00"
        assert property_value["effectiveUntil"] == "9999-12-31T23:59:59.9999999+00:00"

        expected_body = {
            "key": {"$ref": property_def.id},
            "value": {
                "labelValue": "My fund configuration property value"
            },
            "effectiveFrom": "0001-01-01T00:00:00.0000000+00:00",
            "effectiveUntil": "9999-12-31T23:59:59.9999999+00:00"
        }
        assert property_value == expected_body

    def test_undump_properties_with_fund_configuration(self, property_def: pd.DefinitionResource):
        dump = {
            "scope": "scope-fund-configuration",
            "code": "code-fund-configuration",
            "display_name": "fund configuration display name",
            "description": "fund configuration description",
            "dealing_filters": [
                {"filter_id": "SUB", "filter": "account.code startswith '3001'"},
                {"filter_id": "RED", "filter": "account.code startswith '3002'"}
            ],
            "pnl_filters": [
                {"filter_id": "SUB", "filter": "account.code startswith '3001'"},
                {"filter_id": "RED", "filter": "account.code startswith '3002'"}
            ],
            "back_out_filters": [
                {"filter_id": "SUB", "filter": "account.code startswith '3001'"},
                {"filter_id": "RED", "filter": "account.code startswith '3002'"}
            ],
            "external_fee_filters": [
                {
                    "filter_id": "ShareClassAPnLAdjustment",
                    "filter": "account.code eq 'ShareClassAPnl'",
                    "applies_to": "PnLBucket"
                }
            ],
            "properties": [
                {
                    "key": {"$ref": property_def.id},
                    "value": {
                        "label_value": "My fund configuration property value",
                        "metric_value": None,
                        "label_set_value": None
                    },
                    "effective_from": "0001-01-01T00:00:00.0000000+00:00",
                    "effective_until": "9999-12-31T23:59:59.9999999+00:00"
                }
            ]
        }

        sut = fc.FundConfigurationResource.model_validate(dump, context={
            "style": "dump",
            "id": "template-id",
            "$refs": {p.id: p for p in [property_def]}
        })

        expected_body = {
            "scope": "scope-fund-configuration",
            "code": "code-fund-configuration",
            "display_name": "fund configuration display name",
            "description": "fund configuration description",
            "dealing_filters": [
                {"filter_id": "SUB", "filter": "account.code startswith '3001'"},
                {"filter_id": "RED", "filter": "account.code startswith '3002'"}
            ],
            "pnl_filters": [
                {"filter_id": "SUB", "filter": "account.code startswith '3001'"},
                {"filter_id": "RED", "filter": "account.code startswith '3002'"}
            ],
            "back_out_filters": [
                {"filter_id": "SUB", "filter": "account.code startswith '3001'"},
                {"filter_id": "RED", "filter": "account.code startswith '3002'"}
            ],
            "external_fee_filters": [
                {
                    "filter_id": "ShareClassAPnLAdjustment",
                    "filter": "account.code eq 'ShareClassAPnl'",
                    "applies_to": "PnLBucket"
                }
            ],
            "properties": {
                f"FundConfiguration/{property_def.scope}/{property_def.code}": {
                    "key": f"FundConfiguration/{property_def.scope}/{property_def.code}",
                    "value": {
                        "label_value": "My fund configuration property value",
                    },
                    "effective_from": "0001-01-01T00:00:00.0000000+00:00",
                    "effective_until": "9999-12-31T23:59:59.9999999+00:00"
                }
            }
        }

        actual_body = sut.model_dump(exclude_none=True, by_alias=False)

        assert actual_body == expected_body



from fbnconfig import Deployment, datatype, property


def configure(env):
    deployment_name = getattr(env, "name", "property")

    df = property.DefinitionRef(
        id="property/dfdomccy", domain=property.Domain.Holding, scope="default", code="DfDomCcy"
    )
    nominal = property.DefinitionRef(
        id="property/nominal", domain=property.Domain.Holding, scope="default", code="Nominal"
    )
    rating = property.DefinitionResource(
        id="rating",
        domain=property.Domain.Instrument,
        scope=deployment_name,
        code="rating",
        display_name="robtest rating ",
        data_type_id=datatype.DataTypeRef(id="default_str", scope="system", code="string"),
        constraint_style=property.ConstraintStyle.Collection,
        property_description="robTest property",
        life_time=property.LifeTime.Perpetual,
        collection_type=property.CollectionType.Array,
    )
    pd = property.DefinitionResource(
        id="pd1",
        domain=property.Domain.Instrument,
        scope=deployment_name,
        code="pd1",
        display_name="robtest pd ",
        data_type_id=property.ResourceId(scope="system", code="number"),
        constraint_style=property.ConstraintStyle.Property,
        property_description="robTest property",
        life_time=property.LifeTime.Perpetual,
        collection_type=None,
    )
    pv_nominal = property.DefinitionResource(
        id="derived",
        domain=property.Domain.Holding,
        data_type_id=property.ResourceId(scope="system", code="number"),
        scope=deployment_name,
        code="PVNominal",
        property_description="nominal_x_df",
        display_name="DF Nominal",
        derivation_formula=property.Formula("{x} * {y}", x=df, y=nominal),
        is_filterable=False,
    )
    more_derived = property.DefinitionResource(
        id="derived_more",
        domain=property.Domain.Holding,
        data_type_id=property.ResourceId(scope="system", code="number"),
        scope=deployment_name,
        code="more_derived",
        property_description="pd1 x df x nominal",
        display_name="DF Nominal pd1",
        derivation_formula=property.Formula("{x} * {y}", x=pv_nominal, y=pd),
        is_filterable=True,
    )

    return Deployment(deployment_name, [pd, df, nominal, pv_nominal, more_derived, rating])


if __name__ == "__main__":
    import os

    import click

    import fbnconfig

    @click.command()
    @click.argument("lusid_url", type=str)
    @click.option("-v", "--vars_file", type=click.File("r"))
    def cli(lusid_url, vars_file):
        host_vars = fbnconfig.load_vars(vars_file)
        d = configure(host_vars)
        fbnconfig.deploy(d, lusid_url, os.environ["FBN_ACCESS_TOKEN"])

    cli()



import datetime as dt
import json
from hashlib import sha256
from types import SimpleNamespace
from zoneinfo import ZoneInfo

import httpx
import pytest

from fbnconfig import fund_accounting, property

TEST_BASE = "https://foo.lusid.com"


def response_hook(response: httpx.Response):
    response.read()
    response.raise_for_status()


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeChartOfAccountsRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_attach_when_present(self, respx_mock):
        # given that the remote definition exists
        respx_mock.get("/api/api/chartofaccounts/testScope/testCode").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        sut = fund_accounting.ChartOfAccountsRef(
            id="chart_example",
            code="testCode",
            scope="testScope"
        )

        # when we call attach
        sut.attach(client)
        # then a get request was made and no exception raised

    def test_attach_when_missing(self, respx_mock):
        # given the remote does not exist
        respx_mock.get("/api/api/chartofaccounts/testScope/testCode").mock(
            return_value=httpx.Response(404, json={})
        )
        client = self.client
        sut = fund_accounting.ChartOfAccountsRef(
            id="chart_example",
            code="testCode",
            scope="testScope"
        )
        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(client)
        assert "Chart of Accounts testScope/testCode not found" in str(ex.value)

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get("/api/api/chartofaccounts/testScope/testCode").mock(
            return_value=httpx.Response(500, json={})
        )
        client = self.client
        sut = fund_accounting.ChartOfAccountsRef(
            id="chart_example",
            code="testCode",
            scope="testScope"
        )
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeChartOfAccounts:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @pytest.fixture
    def stock_chart_of_account_resource(self):
        property_example = property.DefinitionRef(
            id="property_id", domain=property.Domain.ChartOfAccounts, scope="sc1", code="cd4"
        )

        perp_prop = fund_accounting.PropertyValue(
            property_key=property_example,
            label_value="Hello"
        )

        return fund_accounting.ChartOfAccountsResource(
            id="exampleId",
            scope="testScope",
            code="testCode",
            display_name="example_display_name",
            description="example_description",
            properties=[perp_prop]
        )

    def test_read(self, respx_mock, stock_chart_of_account_resource):
        respx_mock.get("/api/api/chartofaccounts/oldScope/oldCode/properties").mock(
            return_value=httpx.Response(200, json={},))
        respx_mock.get("/api/api/chartofaccounts/oldScope/oldCode").mock(
            return_value=httpx.Response(
                200,
                json={
                    "id": {
                        "scope": "oldScope",
                        "code": "oldCode"
                    },
                    "displayName": "different_example_display_name",
                    "description": "different_example_description",
                    "properties": {}
                    },
            )
        )
        sut = stock_chart_of_account_resource
        old_state = SimpleNamespace(scope="oldScope", code="oldCode")
        result = sut.read(self.client, old_state)
        assert result is not None
        assert result["id"]["scope"] == "oldScope"
        assert result["id"]["code"] == "oldCode"

    def test_create_without_properties(self, respx_mock):
        respx_mock.post("/api/api/chartofaccounts/testScope").mock(
            return_value=httpx.Response(200, json={
                "version": {
                    "effectiveFrom": "2019-08-24T14:15:22Z",
                    "asAtDate": "2019-08-24T14:15:22Z",
                    "asAtCreated": "2019-08-24T14:15:22Z",
                    "userIdCreated": "string",
                    "requestIdCreated": "string",
                    "reasonCreated": "string",
                    "asAtModified": "2019-08-24T14:15:22Z",
                    "userIdModified": "string",
                    "requestIdModified": "string",
                    "reasonModified": "string",
                    "asAtVersionNumber": -2147483648,
                    "entityUniqueId": "string",
                    "stagedModificationIdModified": "string"
                },
            }))

        sut = fund_accounting.ChartOfAccountsResource(
            id="exampleId",
            scope="testScope",
            code="testCode",
            display_name="example_display_name",
            description="example_description",
        )

        result = sut.create(self.client)

        assert result is not None

        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/chartofaccounts/testScope"

        body = json.loads(request.content)

        assert body == {
            "code": "testCode",
            "displayName": "example_display_name",
            "description": "example_description",
        }

        # To get the expected source version
        source_version = sut.__get_content_hash__()

        assert result == {
            "scope": "testScope",
            "code": "testCode",
            "source_version": source_version,
            "remote_version": -2147483648
        }

    def test_create_with_properties(self, respx_mock, stock_chart_of_account_resource):
        respx_mock.post("/api/api/chartofaccounts/testScope").mock(
            return_value=httpx.Response(200, json={
                "version": {
                    "effectiveFrom": "2019-08-24T14:15:22Z",
                    "asAtDate": "2019-08-24T14:15:22Z",
                    "asAtCreated": "2019-08-24T14:15:22Z",
                    "userIdCreated": "string",
                    "requestIdCreated": "string",
                    "reasonCreated": "string",
                    "asAtModified": "2019-08-24T14:15:22Z",
                    "userIdModified": "string",
                    "requestIdModified": "string",
                    "reasonModified": "string",
                    "asAtVersionNumber": -2147483648,
                    "entityUniqueId": "string",
                    "stagedModificationIdModified": "string"
                },
            }))

        sut = stock_chart_of_account_resource
        result = sut.create(self.client)

        assert result is not None

        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/chartofaccounts/testScope"

        body = json.loads(request.content)

        assert body == {
            "code": "testCode",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": {
                "ChartOfAccounts/sc1/cd4": {
                    "key": "ChartOfAccounts/sc1/cd4",
                    "value": {
                        "labelValue": "Hello"
                    },
                }
            }
        }

        # To get the expected source version
        source_version = sut.__get_content_hash__()

        assert result == {
            "scope": "testScope",
            "code": "testCode",
            "source_version": source_version,
            "remote_version": -2147483648
        }

    def test_create_with_prop_time_var(self, respx_mock, stock_chart_of_account_resource):
        respx_mock.post("/api/api/chartofaccounts/testScope").mock(
            return_value=httpx.Response(200, json={
                "version": {
                    "effectiveFrom": "2019-08-24T14:15:22Z",
                    "asAtDate": "2019-08-24T14:15:22Z",
                    "asAtCreated": "2019-08-24T14:15:22Z",
                    "userIdCreated": "string",
                    "requestIdCreated": "string",
                    "reasonCreated": "string",
                    "asAtModified": "2019-08-24T14:15:22Z",
                    "userIdModified": "string",
                    "requestIdModified": "string",
                    "reasonModified": "string",
                    "asAtVersionNumber": -2147483648,
                    "entityUniqueId": "string",
                    "stagedModificationIdModified": "string"
                },
            }))

        property_example = property.DefinitionRef(
            id="property_id", domain=property.Domain.ChartOfAccounts, scope="sc1", code="cd4"
        )

        effective_from = dt.datetime(2000, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))
        effective_until = dt.datetime(2030, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))

        time_prop = fund_accounting.PropertyValue(
            property_key=property_example,
            label_value="Hello",
            effective_from=effective_from,
            effective_until=effective_until
        )

        sut = fund_accounting.ChartOfAccountsResource(
            id="exampleId",
            scope="testScope",
            code="testCode",
            display_name="example_display_name",
            description="example_description",
            properties=[time_prop]
        )

        result = sut.create(self.client)

        assert result is not None

        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/chartofaccounts/testScope"

        body = json.loads(request.content)

        assert body == {
            "code": "testCode",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": {
                "ChartOfAccounts/sc1/cd4": {
                    "key": "ChartOfAccounts/sc1/cd4",
                    "value": {
                        "labelValue": "Hello"
                    },
                    "effectiveFrom": "2000-10-02T13:30:45Z",
                    "effectiveUntil": "2030-10-02T13:30:45Z"
                },
            }
        }

        # To get the expected source version
        source_version = sut.__get_content_hash__()

        assert result == {
            "scope": "testScope",
            "code": "testCode",
            "source_version": source_version,
            "remote_version": -2147483648
        }

    def test_parse_api(self):
        property_example = property.DefinitionRef(
            id="property_id", domain=property.Domain.ChartOfAccounts, scope="sc1", code="cd4"
        )
        data = {
            "scope": "fakeScope",  # note no actual scope in the body of the api response
            "code": "testCode",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": {
                "ChartOfAccounts/sc1/cd4": {
                    "key": {"$ref": "property_id"},
                    "value": {
                        "labelValue": "Hello"
                    },
                }
            }
        }
        # when we parse it
        sut = fund_accounting.ChartOfAccountsResource.model_validate(data, context={
            "id": "something",
            "$refs": {
                "property_id": property_example
            }
        })
        # then properties is an array containing the property_example
        assert sut.properties
        assert len(sut.properties) == 1
        assert sut.properties[0].property_key == property_example
        assert sut.properties[0].label_value == "Hello"

    def test_update_no_change(self, respx_mock, stock_chart_of_account_resource):
        respx_mock.get("/api/api/chartofaccounts/testScope/testCode/properties").mock(
            return_value=httpx.Response(200, json={},))
        sut = stock_chart_of_account_resource

        remote_response = {
            "id": {
                "scope": "testScope",
                "code": "testCode"
            },
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": {
                "ChartOfAccounts/sc1/cd4": {
                    "key": "ChartOfAccounts/sc1/cd4",
                    "value": {
                        "labelValue": "Hello"
                    },
                }
            },
            "version": {
                "effectiveFrom": "2019-08-24T14:15:22Z",
                "asAtDate": "2019-08-24T14:15:22Z",
                "asAtCreated": "2019-08-24T14:15:22Z",
                "userIdCreated": "string",
                "requestIdCreated": "string",
                "reasonCreated": "string",
                "asAtModified": "2019-08-24T14:15:22Z",
                "userIdModified": "string",
                "requestIdModified": "string",
                "reasonModified": "string",
                "asAtVersionNumber": -2147483648,
                "entityUniqueId": "string",
                "stagedModificationIdModified": "string"
            },
        }

        respx_mock.get("/api/api/chartofaccounts/testScope/testCode").mock(
            return_value=httpx.Response(200, json=remote_response)
        )

        # Calculate hashes to simulate no change scenario
        source_hash = sut.__get_content_hash__()

        old_state = SimpleNamespace(
            scope="testScope",
            code="testCode",
            source_version=source_hash,
            remote_version=-2147483648
        )
        result = sut.update(self.client, old_state)
        assert result is None

    def test_update_with_change(self, respx_mock):
        respx_mock.get("/api/api/chartofaccounts/testScope/testCode/properties").mock(
            return_value=httpx.Response(200, json={},))
        respx_mock.get("/api/api/chartofaccounts/testScope/testCode").mock(
            return_value=httpx.Response(
                200,
                json={
                    "id": {
                        "scope": "testScope",
                        "code": "testCode"
                    },
                    "displayName": "example_display_name",
                    "description": "example_description",
                    "properties": {
                        "ChartOfAccounts/sc1/cd4": {
                            "key": "ChartOfAccounts/sc1/cd4",
                            "value": {
                                "labelValue": "Hello"
                            },
                        }
                    },
                    "version": {
                        "effectiveFrom": "2019-08-24T14:15:22Z",
                        "asAtDate": "2019-08-24T14:15:22Z",
                        "asAtCreated": "2019-08-24T14:15:22Z",
                        "userIdCreated": "string",
                        "requestIdCreated": "string",
                        "reasonCreated": "string",
                        "asAtModified": "2019-08-24T14:15:22Z",
                        "userIdModified": "string",
                        "requestIdModified": "string",
                        "reasonModified": "string",
                        "asAtVersionNumber": -2147483648,
                        "entityUniqueId": "string",
                        "stagedModificationIdModified": "string"
                    },
                },
            )
        )
        respx_mock.delete("/api/api/chartofaccounts/testScope/testCode").mock(
            return_value=httpx.Response(200, json={})
        )
        respx_mock.post("/api/api/chartofaccounts/testScope").mock(
            return_value=httpx.Response(200, json={
                "version": {
                    "effectiveFrom": "2019-08-24T14:15:22Z",
                    "asAtDate": "2019-08-24T14:15:22Z",
                    "asAtCreated": "2019-08-24T14:15:22Z",
                    "userIdCreated": "string",
                    "requestIdCreated": "string",
                    "reasonCreated": "string",
                    "asAtModified": "2019-08-24T14:15:22Z",
                    "userIdModified": "string",
                    "requestIdModified": "string",
                    "reasonModified": "string",
                    "asAtVersionNumber": -123456789,
                    "entityUniqueId": "string",
                    "stagedModificationIdModified": "string"
                },
            }))

        old_state = SimpleNamespace(
            scope="testScope", code="testCode", remote_version="oldhash", source_version="different_hash"
        )

        updated_chart_of_account = fund_accounting.ChartOfAccountsResource(
            id="newId",
            scope="testScope",
            code="testCode",
            display_name="new_display_name",
            description="new_description",
        )

        result = updated_chart_of_account.update(self.client, old_state)
        assert result is not None

        # To get the expected source version
        source_version = updated_chart_of_account.__get_content_hash__()

        assert result == {
            "scope": "testScope",
            "code": "testCode",
            "source_version": source_version,
            "remote_version": -123456789
        }

        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/chartofaccounts/testScope"

        # Body is now the updated version
        assert json.loads(request.content) == {
            "code": "testCode",
            "displayName": "new_display_name",
            "description": "new_description",
        }

        # Verify delete and create were called
        assert len(respx_mock.calls) == 4  # read (which also calls a get), delete, create
        assert respx_mock.calls[2].request.method == "DELETE"
        assert respx_mock.calls[3].request.method == "POST"

    def test_delete(self, respx_mock):
        respx_mock.delete(
            "api/api/chartofaccounts/testScope/testCode"
        ).mock(return_value=httpx.Response(200))
        client = self.client
        old_state = SimpleNamespace(scope="testScope", code="testCode")
        fund_accounting.ChartOfAccountsResource.delete(client, old_state)
        assert respx_mock.calls.last.request.method == "DELETE"

    def test_deps(self, stock_chart_of_account_resource):
        sut = stock_chart_of_account_resource
        property_example = property.DefinitionRef(
            id="property_id", domain=property.Domain.ChartOfAccounts, scope="sc1", code="cd4"
        )
        assert sut.deps() == [property_example]

    def test_dump(self):
        property_example = property.DefinitionRef(
            id="property_id", domain=property.Domain.ChartOfAccounts, scope="sc1", code="cd4"
        )

        effective_from = dt.datetime(2000, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))
        effective_until = dt.datetime(2030, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))

        time_prop = fund_accounting.PropertyValue(
            property_key=property_example,
            label_value="Hello",
            effective_from=effective_from,
            effective_until=effective_until
        )

        sut = fund_accounting.ChartOfAccountsResource(
            id="exampleId",
            scope="testScope",
            code="testCode",
            display_name="example_display_name",
            description="example_description",
            properties=[time_prop]
        )

        # when we dump it
        result = sut.model_dump(
            mode="json", by_alias=True, round_trip=True, exclude_none=True, context={"style": "dump"}
        )

        assert result == {
            "scope": "testScope",
            "code": "testCode",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": [
                {
                    "propertyKey": {
                        "$ref": "property_id"
                    },
                    "labelValue": "Hello",
                    "effectiveFrom": "2000-10-02T13:30:45+00:00",
                    "effectiveUntil": "2030-10-02T13:30:45+00:00"
                }
            ]
        }

    def test_undump(self):
        # given dump data with $ref values
        prop1 = property.DefinitionRef(
            id="property_id", domain=property.Domain.Transaction, scope="sc1", code="TestProp"
        )

        effective_from = dt.datetime(2000, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))
        effective_until = dt.datetime(2030, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))

        data = {
            "scope": "testScope",
            "code": "testCode",
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": [
                {
                    "propertyKey": {
                        "$ref": "property_id"
                    },
                    "labelValue": "Test Label",
                    "effectiveFrom": "2000-10-02T13:30:45+00:00",
                    "effectiveUntil": "2030-10-02T13:30:45+00:00"
                }
            ]
        }

        result = fund_accounting.ChartOfAccountsResource.model_validate(
            data,
            context={
                "style": "undump",
                "$refs": {
                    "property_id": prop1
                },
                "id": "undump_chart",
            }
        )

        assert result.code == "testCode"
        assert result.display_name == "example_display_name"
        assert result.description == "example_description"
        assert result.properties
        assert len(result.properties) == 1
        assert result.properties[0].property_key == prop1
        assert result.properties[0].property_key.code == "TestProp"
        assert result.properties[0].label_value == "Test Label"
        assert result.properties[0].effective_from == effective_from
        assert result.properties[0].effective_until == effective_until

    def test_parse_api_format(self):
        # given dump data with $ref values
        prop1 = property.DefinitionRef(
            id="property_id", domain=property.Domain.Transaction, scope="sc1", code="TestProp"
        )

        effective_from = dt.datetime(2000, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))
        effective_until = dt.datetime(2030, 10, 2, 14, 30, 45, tzinfo=ZoneInfo("Europe/London"))

        prop2 = property.DefinitionRef(
            id="property_id_2", domain=property.Domain.Transaction, scope="sc2", code="TestProp2"
        )

        data = {
            "id": {
                "scope": "testScope",
                "code": "testCode"
            },
            "displayName": "example_display_name",
            "description": "example_description",
            "properties": {
                "ChartOfAccounts/sc1/cd4": {
                    "key": {
                        "$ref": "property_id"
                    },
                    "labelValue": "Test Label",
                    "effectiveFrom": "2000-10-02T13:30:45+00:00",
                    "effectiveUntil": "2030-10-02T13:30:45+00:00"

                },
                "ChartOfAccounts/sc2/cd2": {
                    "key": {
                        "$ref": "property_id_2"
                    },
                    "labelValue": "Test Label 2"
                }
            }
        }

        result = fund_accounting.ChartOfAccountsResource.model_validate(
            data,
            context={
                "style": "api",
                "$refs": {
                    "property_id": prop1,
                    "property_id_2": prop2
                },
                "id": "undump_chart",
            }
        )

        assert result.scope == "testScope"
        assert result.code == "testCode"
        assert result.display_name == "example_display_name"
        assert result.description == "example_description"
        assert result.properties
        assert len(result.properties) == 2
        assert result.properties[0].property_key == prop1
        assert result.properties[0].property_key.code == "TestProp"
        assert result.properties[0].label_value == "Test Label"
        assert result.properties[0].effective_from == effective_from
        assert result.properties[0].effective_until == effective_until
        assert result.properties[1].property_key == prop2
        assert result.properties[1].property_key.code == "TestProp2"
        assert result.properties[1].label_value == "Test Label 2"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeAccountsRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_attach_when_present(self, respx_mock):
        # given that the remote definition exists
        respx_mock.get("/api/api/chartofaccounts/testScope/testCode/accounts/accountCode").mock(
            return_value=httpx.Response(200, json={})
        )
        client = self.client
        chart = fund_accounting.ChartOfAccountsRef(
            id="chart_example",
            code="testCode",
            scope="testScope"
        )
        sut = fund_accounting.AccountRef(
            id="account_example",
            account_code="accountCode",
            chart_of_accounts=chart,
        )

        # when we call attach
        sut.attach(client)
        # then a get request was made and no exception raised

    def test_attach_when_missing(self, respx_mock):
        # given the remote does not exist
        respx_mock.get("/api/api/chartofaccounts/testScope/testCode/accounts/accountCode").mock(
            return_value=httpx.Response(404, json={})
        )
        client = self.client
        chart = fund_accounting.ChartOfAccountsRef(
            id="chart_example",
            code="testCode",
            scope="testScope"
        )
        sut = fund_accounting.AccountRef(
            id="account_example",
            account_code="accountCode",
            chart_of_accounts=chart,
        )
        # when we call attach an exception is raised
        with pytest.raises(RuntimeError) as ex:
            sut.attach(client)
        assert "Account accountCode not found" in str(ex.value)

    def test_attach_when_http_error(self, respx_mock):
        # given the server returns an error
        respx_mock.get("/api/api/chartofaccounts/testScope/testCode/accounts/accountCode").mock(
            return_value=httpx.Response(500, json={})
        )
        client = self.client
        chart = fund_accounting.ChartOfAccountsRef(
            id="chart_example",
            code="testCode",
            scope="testScope"
        )
        sut = fund_accounting.AccountRef(
            id="account_example",
            account_code="accountCode",
            chart_of_accounts=chart,
        )
        # when we call attach an exception is raised
        with pytest.raises(httpx.HTTPError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeAccounts:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @pytest.fixture
    def stock_account_resource(self):
        chart = fund_accounting.ChartOfAccountsResource(
            id="exampleId",
            scope="testScope",
            code="testCode",
            display_name="example_display_name",
            description="example_description",
        )

        perp_prop = fund_accounting.PropertyValue(
            property_key=property.DefinitionRef(
                id="one", domain=property.Domain.Account, scope="sc1", code="cd4"
            ),
            label_value="Hello"
        )

        return fund_accounting.AccountResource(
            id="example_id",
            chart_of_accounts=chart,
            code="account_code",
            description="example_desc",
            type=fund_accounting.AccountType.ASSET,
            status=fund_accounting.AccountStatus.ACTIVE,
            control="Manual",
            properties=[perp_prop]
        )

    def test_read(self, respx_mock, stock_account_resource):
        respx_mock.get("/api/api/chartofaccounts/oldScope/oldCode/accounts/different_account_code/properties").mock(
            return_value=httpx.Response(200, json={},))
        respx_mock.get("/api/api/chartofaccounts/oldScope/oldCode/accounts/different_account_code").mock(
            return_value=httpx.Response(
                200,
                json={
                    "code": "different_account_code",
                    "description": "different_example_desc",
                    "type": "Asset",
                    "status": "Active",
                    "control": "Manual",
                    "properties": {
                        "ChartOfAccounts/sc1/cd4": {
                            "key": "ChartOfAccounts/sc1/cd4",
                            "value": {
                                "labelValue": "Hello"
                            },
                        }
                    },
                }

            )
        )
        sut = stock_account_resource
        old_state = SimpleNamespace(
            scope="oldScope",
            code="oldCode",
            account_code="different_account_code"
        )
        result = sut.read(self.client, old_state)
        assert result is not None
        assert result["code"] == "different_account_code"

    def test_create_without_properties(self, respx_mock):
        response = {
            "accounts": [
                {
                    "code": "account_code",
                    "description": "example_desc",
                    "type": "Asset",
                    "status": "Active",
                    "control": "Manual",
                    "properties": {
                        "Account/sc1/cd1": {
                            "key": "Account/sc1/cd1",
                            "value": {
                                "labelValue": "Hello"
                            },
                        }
                    }
                }
            ]
        }

        respx_mock.post("/api/api/chartofaccounts/testScope/testCode/accounts").mock(
            return_value=httpx.Response(200, json=response))

        chart = fund_accounting.ChartOfAccountsResource(
            id="exampleId",
            scope="testScope",
            code="testCode",
            display_name="example_display_name",
            description="example_description",
        )

        sut = fund_accounting.AccountResource(
            id="example_id",
            chart_of_accounts=chart,
            code="account_code",
            description="example_desc",
            type=fund_accounting.AccountType.ASSET,
            status=fund_accounting.AccountStatus.ACTIVE,
            control="Manual",
        )

        result = sut.create(self.client)

        assert result is not None

        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/chartofaccounts/testScope/testCode/accounts"

        body = json.loads(request.content)
        assert body == [
            {
                "code": "account_code",
                "description": "example_desc",
                "type": "Asset",
                "status": "Active",
                "control": "Manual"
            },
        ]

        dump = sut.model_dump(mode="json", exclude_none=True, by_alias=True)
        source_version = sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()
        account_to_dump = response["accounts"][0]

        remote_version = sha256(json.dumps(account_to_dump, sort_keys=True).encode()).hexdigest()

        assert result == {
            "account_code": "account_code",
            "scope": "testScope",
            "code": "testCode",
            "source_version": source_version,
            "remote_version": remote_version,
        }

    def test_create_with_properties(self, respx_mock, stock_account_resource):
        response = {
            "accounts": [
                {
                    "code": "account_code",
                    "description": "example_desc",
                    "type": "Asset",
                    "status": "Active",
                    "control": "Manual",
                    "properties": {
                        "Account/sc1/cd1": {
                        "key": "Account/sc1/cd1",
                        "value": {
                            "labelValue": "Hello"
                        },
                        }
                    }
                }
            ]
        }
        respx_mock.post("/api/api/chartofaccounts/testScope/testCode/accounts").mock(
            return_value=httpx.Response(200, json=response))

        sut = stock_account_resource
        result = sut.create(self.client)

        assert result is not None

        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/chartofaccounts/testScope/testCode/accounts"

        body = json.loads(request.content)

        assert body == [
            {
                "code": "account_code",
                "description": "example_desc",
                "type": "Asset",
                "status": "Active",
                "control": "Manual",
                "properties": {
                    "Account/sc1/cd4": {
                        "key": "Account/sc1/cd4",
                        "value": {
                            "labelValue": "Hello"
                        },
                    }
                }
            }
        ]

        dump = sut.model_dump(mode="json", exclude_none=True, by_alias=True)
        source_version = sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()
        account_to_dump = response["accounts"][0]

        remote_version = sha256(json.dumps(account_to_dump, sort_keys=True).encode()).hexdigest()

        assert result == {
            "account_code": "account_code",
            "scope": "testScope",
            "code": "testCode",
            "source_version": source_version,
            "remote_version": remote_version,
        }

    def test_update_no_change(self, respx_mock, stock_account_resource):
        respx_mock.get("/api/api/chartofaccounts/testScope/testCode/accounts/account_code/properties").mock(
            return_value=httpx.Response(200, json={},))
        sut = stock_account_resource

        remote_response = {
            "code": "account_code",
            "description": "example_desc",
            "type": "Asset",
            "status": "Active",
            "control": "Manual",
            "properties": {
                "ChartOfAccounts/sc1/cd4": {
                    "key": "ChartOfAccounts/sc1/cd4",
                    "value": {
                        "labelValue": "Hello"
                    },
                }
            },
        }

        respx_mock.get("/api/api/chartofaccounts/testScope/testCode/accounts/account_code").mock(
            return_value=httpx.Response(200, json=remote_response)
        )

        # Calculate hashes to simulate no change scenario
        source_hash = sut.__get_content_hash__()

        # The remote hash should match what update() calculates from the read() response
        remote_hash = sha256(json.dumps(remote_response, sort_keys=True).encode()).hexdigest()

        old_state = SimpleNamespace(
            account_code="account_code",
            scope="testScope",
            code="testCode",
            source_version=source_hash,
            remote_version=remote_hash
        )
        result = sut.update(self.client, old_state)
        assert result is None

    def test_update_with_change(self, respx_mock):
        respx_mock.get("/api/api/chartofaccounts/testScope/testCode/accounts/account_code/properties").mock(
            return_value=httpx.Response(200, json={},))
        remote_response = {
                "code": "account_code",
                "description": "example_desc",
                "type": "Asset",
                "status": "Active",
                "control": "Manual",
                "properties": {
                    "ChartOfAccounts/sc1/cd4": {
                        "key": "ChartOfAccounts/sc1/cd4",
                        "value": {
                            "labelValue": "Hello"
                        },
                    }
                },
            }

        respx_mock.get("/api/api/chartofaccounts/testScope/testCode/accounts/account_code").mock(
            return_value=httpx.Response(200, json=remote_response)
        )

        response = {
            "accounts": [
                {
                    "code": "account_code",
                    "description": "different_example_desc",
                    "type": "Asset",
                    "status": "Active",
                    "control": "Manual",
                    "properties": {
                        "Account/sc2/cd2": {
                            "key": "Account/sc2/cd2",
                            "value": {
                                "labelValue": "Goodbye"
                            },
                        }
                    }
                }
            ]
        }

        respx_mock.post("/api/api/chartofaccounts/testScope/testCode/accounts").mock(
            return_value=httpx.Response(200, json=response))

        old_state = SimpleNamespace(
            account_code="account_code",
            scope="testScope",
            code="testCode",
            remote_version="oldhash",
            source_version="different_hash"
        )

        chart = fund_accounting.ChartOfAccountsResource(
            id="exampleId",
            scope="testScope",
            code="testCode",
            display_name="example_display_name",
            description="example_description",
        )

        different_perp_prop = fund_accounting.PropertyValue(
            property_key=property.DefinitionRef(
                id="one", domain=property.Domain.Account, scope="sc2", code="cd2"
            ),
            label_value="Goodbye"
        )

        updated_account = fund_accounting.AccountResource(
            id="different_example_id",
            chart_of_accounts=chart,
            code="account_code",
            description="different_example_desc",
            type=fund_accounting.AccountType.ASSET,
            status=fund_accounting.AccountStatus.ACTIVE,
            control="Manual",
            properties=[different_perp_prop]
        )

        result = updated_account.update(self.client, old_state)
        assert result is not None

        dump = updated_account.model_dump(mode="json", exclude_none=True, by_alias=True)
        source_version = sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()
        account_to_dump = response["accounts"][0]

        remote_version = sha256(json.dumps(account_to_dump, sort_keys=True).encode()).hexdigest()

        assert result == {
            "account_code": "account_code",
            "scope": "testScope",
            "code": "testCode",
            "source_version": source_version,
            "remote_version": remote_version,
        }

        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/chartofaccounts/testScope/testCode/accounts"

        body = json.loads(request.content)

        # Body is now the updated version
        assert body == [
            {
                "code": "account_code",
                "description": "different_example_desc",
                "type": "Asset",
                "status": "Active",
                "control": "Manual",
                "properties": {
                    "Account/sc2/cd2": {
                        "key": "Account/sc2/cd2",
                        "value": {
                            "labelValue": "Goodbye"
                        },
                    }
                }
            }
        ]

        # Verify upsert was called
        assert len(respx_mock.calls) == 3
        assert respx_mock.calls[2].request.method == "POST"

    def test_cannot_update_if_scope_changes(self, stock_account_resource):
        old_state = SimpleNamespace(
            account_code="account_code",
            scope="different_scope",
            code="testCode",
            remote_version="oldhash",
            source_version="different_hash"
        )

        sut = stock_account_resource
        error_message = "Cannot change the scope on an accounts"
        with pytest.raises(RuntimeError, match=error_message):
            sut.update(self.client, old_state)

    def test_cannot_update_if_code_changes(self, stock_account_resource):
        old_state = SimpleNamespace(
            account_code="account_code",
            scope="testScope",
            code="different_code",
            remote_version="oldhash",
            source_version="different_hash"
        )

        sut = stock_account_resource
        error_message = "Cannot change the code on an account"
        with pytest.raises(RuntimeError, match=error_message):
            sut.update(self.client, old_state)

    def test_delete(self, respx_mock):
        respx_mock.post(
            "/api/api/chartofaccounts/testScope/testCode/accounts/$delete"
        ).mock(return_value=httpx.Response(200))
        client = self.client
        old_state = SimpleNamespace(account_code="account_code", scope="testScope", code="testCode")
        fund_accounting.AccountResource.delete(client, old_state)
        assert respx_mock.calls.last.request.method == "POST"

    def test_deps_without_properties(self):
        chart = fund_accounting.ChartOfAccountsResource(
            id="chartDepend",
            scope="testScope",
            code="testCode",
            display_name="example_display_name",
            description="example_description",
        )

        account = fund_accounting.AccountResource(
            id="example_id",
            chart_of_accounts=chart,
            code="account_code",
            description="example_desc",
            type=fund_accounting.AccountType.ASSET,
            status=fund_accounting.AccountStatus.ACTIVE,
            control="Manual",
        )
        assert account.deps() == [chart]

    def test_deps_with_properties(self, stock_account_resource):
        chart_prop = property.DefinitionRef(
            id="property_id", domain=property.Domain.ChartOfAccounts, scope="sc1", code="cd4"
        )

        chart_perp_prop = fund_accounting.PropertyValue(
            property_key=chart_prop,
            label_value="Hello"
        )

        chart = fund_accounting.ChartOfAccountsResource(
            id="exampleId",
            scope="testScope",
            code="testCode",
            display_name="example_display_name",
            description="example_description",
            properties=[chart_perp_prop]
        )

        account_prop = property.DefinitionRef(
            id="one", domain=property.Domain.Account, scope="sc1", code="cd4"
        )

        account_perp_prop = fund_accounting.PropertyValue(
            property_key=account_prop,
            label_value="Hello"
        )

        sut = fund_accounting.AccountResource(
            id="example_id",
            chart_of_accounts=chart,
            code="account_code",
            description="example_desc",
            type=fund_accounting.AccountType.ASSET,
            status=fund_accounting.AccountStatus.ACTIVE,
            control="Manual",
            properties=[account_perp_prop]
        )

        assert sut.deps() == [chart, chart_prop, account_prop]

    def test_dump(self, stock_account_resource):
        sut = stock_account_resource

        # when we dump it
        result = sut.model_dump(
            mode="json", by_alias=True, round_trip=True, exclude_none=True, context={"style": "dump"}
        )

        assert result == {
            "chartOfAccounts": {
                "$ref": "exampleId"
            },
            "code": "account_code",
            "description": "example_desc",
            "type": "Asset",
            "status": "Active",
            "control": "Manual",
            "properties": [
                {
                    "propertyKey": {
                        "$ref": "one"
                    },
                    "labelValue": "Hello"
                }
            ]
        }

    def test_undump(self):
        chart = fund_accounting.ChartOfAccountsResource(
            id="exampleId",
            scope="testScope",
            code="testCode",
            display_name="example_display_name",
            description="example_description",
        )

        # given dump data with $ref values
        prop1 = property.DefinitionRef(
            id="property_id", domain=property.Domain.Transaction, scope="sc1", code="TestProp"
        )

        data = {
            "chartOfAccounts": {
                "$ref": "exampleId"
            },
            "code": "account_code",
            "description": "example_desc",
            "type": "Asset",
            "status": "Active",
            "control": "Manual",
            "properties": [
                {
                    "propertyKey": {
                        "$ref": "property_id"
                    },
                    "labelValue": "Test Label"
                }
            ]
        }

        result = fund_accounting.AccountResource.model_validate(
            data,
            context={
                "style": "undump",
                "$refs": {
                    "exampleId": chart,
                    "property_id": prop1
                },
                "id": "undump_account",
            }
        )

        assert result.chart_of_accounts == chart
        assert result.code == "account_code"
        assert result.description == "example_desc"
        assert result.properties
        assert len(result.properties) == 1
        assert result.properties[0].property_key == prop1
        assert result.properties[0].property_key.code == "TestProp"
        assert result.properties[0].label_value == "Test Label"



from fbnconfig import Deployment, workspace


def configure(env):
    deployment_name = getattr(env, "name", "workspace_example")
    wksp = workspace.WorkspaceResource(
        id="wk1",
        visibility=workspace.Visibility.SHARED,
        name=deployment_name,
        description="workspace number one",
    )
    item1 = workspace.WorkspaceItemResource(
        id="item1",
        workspace=wksp,
        group="group1",
        name="item1",
        description="item one version two",
        type="lusid-web-dashboard",
        format=1,
        content={"msg": "some text"},
    )
    item2 = workspace.WorkspaceItemResource(
        id="item2",
        workspace=wksp,
        group="group1",
        name="item2",
        description="item one version two",
        format=1,
        type="lusid-web-dashboard",
        content={"msg": "some text", "foo": 10},
    )
    return Deployment(deployment_name, [item1, item2])



from __future__ import annotations

from types import SimpleNamespace
from typing import Annotated, Any, Dict, Optional, Sequence

import httpx
from pydantic import BaseModel, BeforeValidator, Field, PlainSerializer, model_validator

from .property import PropertyKey
from .resource_abc import Ref, Resource, register_resource


@register_resource()
class SideRef(BaseModel, Ref):
    """Used to reference an existing side
    Example
    -------
    >>> from fbnconfig.side_definition import SideRef
    >>> side = SideRef(id="side", scope="scope", side="side")
    """

    id: str = Field(exclude=True)
    scope: str
    side: str

    def attach(self, client):
        scope, side = self.scope, self.side
        try:
            client.get(f"/api/api/transactionconfiguration/sides/{side}", params={"scope": scope})
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Side {self.side} does not exist in scope {self.scope}")
            else:
                raise ex


@register_resource()
class SideResource(BaseModel, Resource):
    """Create a side definition

    Example
    -------
    >>> from fbnconfig.side_definition import SideResource
    >>> side = SideResource(
    >>>     id="side",
    >>>     side="side",
    >>>     scope="scope",
    >>>     security="Txn:LusidInstrumentId",
    >>>     currency = "Txn:TradeCurrency",
    >>>     rate = "Txn:TradeToPortfolioRate",
    >>>     units = "Txn:Units",
    >>>     amount = "Txn:TotalConsideration",
    >>>     notional_amount = "0"
    >>> )
    """

    id: str = Field(exclude=True, init=True)
    side: str
    scope: str
    security: str | PropertyKey
    currency: str | PropertyKey
    rate: str | PropertyKey
    units: str | PropertyKey
    amount: str | PropertyKey
    notional_amount: str | PropertyKey

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    def create(self, client: httpx.Client) -> Optional[Dict[str, Any]]:
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True, exclude={"scope"})
        scope = self.scope
        client.put(
            f"/api/api/transactionconfiguration/sides/{self.side}", json=desired, params={"scope": scope}
        )
        return {"side": self.side, "scope": self.scope}

    def read(self, client: httpx.Client, old_state: SimpleNamespace):
        side = old_state.side
        scope = old_state.scope
        return client.get(
            f"/api/api/transactionconfiguration/sides/{side}", params={"scope": scope}
        ).json()

    def update(self, client: httpx.Client, old_state: SimpleNamespace):
        # Check for scope or name change, must delete and create
        if [old_state.side, old_state.scope] != [self.side, self.scope]:
            self.delete(client, old_state)
            return self.create(client)
        # Has there been a change?
        remote = self.read(client, old_state) or {}
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True, exclude={"scope"})
        effective = remote | desired
        if effective == remote:
            return None
        return self.create(client)

    @staticmethod
    def delete(client: httpx.Client, old_state: SimpleNamespace):
        side, scope = old_state.side, old_state.scope
        client.delete(f"/api/api/transactionconfiguration/sides/{side}/$delete", params={"scope": scope})

    def deps(self) -> Sequence[Resource | Ref]:
        return [
            value
            for value in (
                self.security,
                self.currency,
                self.amount,
                self.rate,
                self.units,
                self.notional_amount,
            )
            if isinstance(value, (Resource, Ref))
        ]


def ser_side_key(value, info):
    if info.context and info.context.get("style") == "dump":
        return {"$ref": value.id}
    return value.side


def des_side_key(value, info):
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get("$refs"):
        ref = info.context["$refs"][value["$ref"]]
        return ref
    return value


SideKey = Annotated[
    SideResource | SideRef,
    BeforeValidator(des_side_key), PlainSerializer(ser_side_key)
]



from __future__ import annotations

import csv
import hashlib
import json
import sys
import textwrap
import time
from datetime import datetime
from enum import StrEnum
from io import StringIO
from types import SimpleNamespace
from typing import Annotated, Any, Dict, List, Sequence, Union

import httpx
from httpx import Client as httpxClient
from pydantic import (
    BaseModel,
    BeforeValidator,
    Field,
    PlainSerializer,
    field_serializer,
    field_validator,
    model_validator,
)

from .backoff_handler import BackoffHandler
from .property import PropertyKey
from .resource_abc import CamelAlias, Ref, Resource, register_resource


# from the sdk
class TaskStatus(StrEnum):
    CREATED = "Created"
    WAITING_FOR_ACTIVATION = "WaitingForActivation"
    WAITING_TO_RUN = "WaitingToRun"
    RUNNING = "Running"
    WAITING_FOR_CHILDREN_TO_COMPLETE = "WaitingForChildrenToComplete"
    RAN_TO_COMPLETION = "RanToCompletion"
    CANCELED = "Canceled"
    FAULTED = "Faulted"


def start_query(client: httpx.Client, sql: str) -> str:
    return client.put(
        "/honeycomb/api/SqlBackground", content=sql, headers={"Content-type": "text/plain"}
    ).json()["executionId"]


def wait_for_background(client, execution_id) -> bool:
    status = get_status(client, execution_id)
    match status:
        case TaskStatus.FAULTED | TaskStatus.CANCELED:
            raise RuntimeError("Query was " + status + ". Execution id: " + execution_id)
        case (
            TaskStatus.CREATED
            | TaskStatus.WAITING_FOR_ACTIVATION
            | TaskStatus.WAITING_TO_RUN
            | TaskStatus.WAITING_FOR_CHILDREN_TO_COMPLETE
        ):
            return False
        case TaskStatus.RUNNING:
            return False
        case TaskStatus.RAN_TO_COMPLETION:
            return True
        case _:
            raise RuntimeError("Unknown status: " + status + ". Execution id: " + execution_id)


def fetch(client, execution_id):
    res = client.get(f"/honeycomb/api/SqlBackground/{execution_id}/jsonproper")
    return res.json()


def get_status(client: httpx.Client, execution_id: str) -> str:
    """Get the status of the query in Luminesce

    Returns:
        str: string containing the query status value.
    """
    return client.get(f"/honeycomb/api/SqlBackground/{execution_id}").json()["status"]


def query(client: httpx.Client, sql: str, backoff_handler: BackoffHandler) -> Dict[Any, Any]:
    execution_id: str = start_query(client, sql)
    progress = False
    while True:
        if progress is False:
            backoff_handler.sleep()
            progress = wait_for_background(client, execution_id)
        else:
            #  Pause after the last wait_for_background_call, otherwise increased chances of 429s
            backoff_handler.sleep()
            return fetch(client, execution_id)


class ParameterType(StrEnum):
    BigInt = "BigInt"
    Boolean = "Boolean"
    Date = "Date"
    DateTime = "DateTime"
    Decimal = "Decimal"
    Double = "Double"
    Int = "Int"
    Table = "Table"
    Text = "Text"


class VariableType(StrEnum):
    Scalar = "@@"
    Table = "@"


class Variable(BaseModel):
    name: str
    type: VariableType
    sql: str

    def init_str(self):
        #  eg @scalar = select 2 + 2
        return f"{self.type.value}{self.name} = {self.sql}"

    def with_str(self):
        return f"{self.type.value}{self.name}"


class Parameter(BaseModel, CamelAlias):
    name: str
    type: ParameterType
    value: Any
    set_as_default_value: bool = True
    is_mandatory: bool = True
    tooltip: str | None = None

    # return in the same format that sys.file stores it in
    def metadata(self):
        base: Dict[str, Any] = {"Name": self.name, "Type": self.type.value, "Description": self.tooltip}
        if self.set_as_default_value and self.type != ParameterType.Table:
            base["DefaultValue"] = self.value
        if self.type == ParameterType.Table:
            if self.is_mandatory:
                base["ConditionUsage"] = 2
            else:
                base["ConditionUsage"] = 0
        return base


def lumi_fmt(value: Any) -> str:
    """ escaping parameters which act like commandline args and use " for strings """
    if value is None:
        return "null"
    if isinstance(value, Variable):
        return f"{value.type.value}{value.name}"
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, (int, float)):
        return str(value)
    return '"' + value.replace('"', '""') + '"'


def lumi_sql(value: Any) -> str:
    """ escape a sql value """
    if value is None:
        return "null"
    if isinstance(value, Variable):
        return f"{value.type.value}{value.name}"
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, (int, float)):
        return str(value)
    if isinstance(value, datetime):
        return "'" + str(value) + "'"
    return "'" + value.replace("'", "''") + "'"


@register_resource()
class ViewRef(BaseModel, Ref):
    """Reference an existing view


    Example
    ----------
    >>> from fbnconfig import lumi
    >>> lumi.ViewRef(
    ...  id="lumi-example-ref",
    ...  provider="Views.fbnconfig.existing_view")


    Attributes
    ----------
    id : str
         Resource identifier.
    provider : str
        Name of the view referenced. This is assumed to exist
    """

    id: str = Field(exclude=True)
    provider: str
    _backoff_handler = BackoffHandler()
    _version: str | None = None

    def attach(self, client) -> None:
        res = query(
            client,
            textwrap.dedent(f"""\
                select r.Version from sys.registration as r
                where name = '{self.provider}'
                order by r.Version asc
                limit 1
            """),
            backoff_handler=self._backoff_handler,
        )
        if len(res) != 1:
            raise RuntimeError(f"Failed to attach ref to {self.provider}, the view might not exist")
        self._version = res[0]


@register_resource()
class ViewResource(BaseModel, Resource):
    """Create and manage a Luminesce view

    Example
    ----------
    >>> from fbnconfig import lumi
    >>> lumi.ViewResource(
    ...  id="lumi-example-view",
    ...  provider="Views.fbnconfig.example",
    ...  description="My resource test view",
    ...  documentation_link="http://example.com/query",
    ...  variable_shape=False,
    ...  use_dry_run=True,
    ...  allow_execute_indirectly=False,
    ...  distinct=True,
    ...  sql='select 2+#PARAMETERVALUE(p1)   as   twelve',
    ...  parameters=[
    ...      lumi.Parameter(
    ...          name="p1",
    ...          value=10,
    ...          set_as_default_value=True,
    ...          tooltip="a number",
    ...          type=lumi.ParameterType.INT
    ...      )

    Attributes
    ----------
    id : str
         Resource identifier.
    provider : str
        Name of the view managed by this resource
    description : str
        View description
    sql: str
        The query string for the view
    parameters : list of `Parameter`, optional
        List of parameters for the view
    dependencies : list of dependencies, optional
        This can be another view or any other resource
    documentation_link: str, optional
        Displays one or more hyperlinks in the summary dialog for the view
    variable_shape: bool, optional
        This is useful if data returned is likely to vary in shape between queries. Defaults to false.
    allow_execute_indirectly : bool, optional
        Allows end users to query providers within the view even if they are not entitled to use those
        providers directly.
        Defaults to false.
    limit: int, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    group_by: str, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    filter: str, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    offset: int, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    distinct: bool, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    use_dry_run: bool, optional
        Intended for automatic deployment of views. See docs for more details. Defaults to false
    variables: List of `Variable`, optional
        A table variable that can be passed into the view by an end user or in code

    See Also
    --------
    `https://support.lusid.com/knowledgebase/article/KA-01767/en-us`__
    """

    id: str = Field(exclude=True)
    provider: str = Field(serialization_alias="Provider")
    description: str = Field(serialization_alias="Description")
    sql: str
    parameters: List[Parameter] = []
    dependencies: List | None = None
    documentation_link: str | None = None
    variable_shape: bool | None = None
    allow_execute_indirectly: bool | None = None
    limit: int | None = None
    group_by: str | None = None
    filter: str | None = None
    offset: int | None = None
    distinct: bool | None = None
    use_dry_run: bool | None = None
    variables: List[Variable] = []

    _backoff_handler = BackoffHandler()

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info):
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    @field_validator("dependencies", mode="before")
    @classmethod
    def des_dependencies(cls, data: Any, info) -> List[Resource | Ref]:
        if data is None:
            return data
        if info.context and info.context.get("$refs"):
            return [
                info.context["$refs"][d["$ref"]] if isinstance(d, dict) else d
                for d in data
            ]
        return data

    class Registration:
        tries = 10
        wait_time = 1

    _saved_options = {  # maps from sys.file metadata to view option names
        "Description": "Description",
        "DocumentationLink": "documentationLink",
        "IsWithinDirectProviderView": "variableShape",
        "IsWithinViewAllowingIndirectExecute": "allowExecuteIndirectly",
    }
    _test_options = ["distinct", "filter", "groupby", "limit", "offset", "preamble", "useDryRun"]

    def read(self, client, old_state) -> Dict[str, Any]:
        path = old_state.provider.replace(".", "/")
        res = query(
            client,
            textwrap.dedent(f"""\
                select f.Content, r.Version from sys.file as f
                join sys.registration as r on r.Name = '{old_state.provider}'
                where path = 'databaseproviders/{path}.sql'
                order by r.Version asc
                limit 1
            """),
            self._backoff_handler,
        )
        # todo: exception here
        assert len(res) == 1

        def strip_column_description(kv: Dict) -> Dict:
            if kv["Type"] == "Table":
                kv["Description"] = kv["Description"].split("\nAvailable columns")[0]

            return kv

        parts = res[0]["Content"].split("--- MetaData ---")
        sql = parts[0]
        metadata = json.loads(parts[1])

        parameters = [strip_column_description(p) for p in metadata["Parameters"]]
        props = {
            v: metadata[k] for k, v in self._saved_options.items() if metadata.get(k, None) is not None
        }
        return {"sql": sql, "version": res[0]["Version"], "parameters": parameters} | props

    @staticmethod
    def registration_version(client, view_name, backoff_handler: BackoffHandler) -> int | None:
        content = textwrap.dedent(f"""\
            select Version from sys.registration where Name='{view_name}'
            order by Version asc
            limit 1
        """)
        rows = query(client, content, backoff_handler)
        return int(rows[0]["Version"]) if len(rows) > 0 else None

    @staticmethod
    def format_option(option, value):
        if isinstance(value, bool) and value:
            return f"--{option}"
        if isinstance(value, (int, float)):
            return f"--{option}={value}"
        # we run self.dump by alias, these two have a serialization alias on the dto
        # which ends up as the option, which should be lower
        if option == "Provider" or option == "Description":
            option = option.lower()
        return f"--{option}={lumi_fmt(value)}"

    def get_variables(self):
        param_variables = [param.value for param in self.parameters if isinstance(param.value, Variable)]
        seen = set()
        return [
            value
            for value in self.variables + param_variables
            if value.name not in seen and not seen.add(value.name)
        ]

    def template(self, desired):
        options = [
            self.format_option(option, desired[option])
            for option in ["Provider"] + self._test_options + list(self._saved_options.values())
            if desired.get(option) is not None
        ]

        tpl = textwrap.dedent("""\
            {preamble}@x = use Sys.Admin.SetupView{with_clause}
            {options}{params}
            ----
            {sql}
            enduse;
            select * from @x;
        """)
        params = [
            f"{p.name},{p.type.value},{lumi_fmt(p.value)},{lumi_fmt(p.is_mandatory)}"
            + (f',"{p.tooltip}"' if p.tooltip is not None else "")
            if p.type == ParameterType.Table
            else f"{p.name},{p.type.value},{lumi_fmt(p.value)},{lumi_fmt(p.set_as_default_value)}"
            + (f',"{p.tooltip}"' if p.tooltip is not None else "")
            for p in self.parameters
        ]
        param_clause = "\n--parameters\n{0}".format("\n".join(params)) if len(params) > 0 else ""
        variables = self.get_variables()
        preamble = ";\n".join([v.init_str() for v in variables]) + ";\n" if len(variables) > 0 else ""
        with_clause = (
            " with " + ", ".join([v.with_str() for v in variables]) if len(variables) > 0 else ""
        )
        sql = tpl.format(
            options="\n".join(options),
            params=param_clause,
            sql=desired["sql"],
            with_clause=with_clause,
            preamble=preamble,
        )
        return sql

    def _get_content_hash(self) -> str:
        desired = self.model_dump(exclude_none=True, by_alias=True)
        sql = self.template(desired)
        return hashlib.sha256(sql.encode()).hexdigest()

    def create(self, client) -> Dict[str, Any]:
        desired = self.model_dump(exclude_none=True, by_alias=True)
        sql = self.template(desired)
        query(client, sql, self._backoff_handler)
        for i in range(1, self.Registration.tries + 1):
            if self.registration_version(client, self.provider, self._backoff_handler) is not None:
                break
            else:
                if i == self.Registration.tries:
                    sys.stderr.write(
                        f"warning: no view registration after {i} tries for {self.provider}"
                    )
                else:
                    time.sleep(self.Registration.wait_time)
        return {"provider": self.provider}

    def update(self, client, old_state):
        if self.provider != old_state.provider:
            self.delete(client, old_state)
            self.create(client)
            return {"provider": self.provider}
        desired = self.model_dump(exclude_none=True, by_alias=True, exclude=set(self._test_options))
        raw_remote = self.read(client, old_state)
        remote_props = {
            k.lower(): v
            for k, v in raw_remote.items()
            if k in self._saved_options.values() and v is not None
        }
        desired_props = {
            k.lower(): v
            for k, v in desired.items()
            if k in self._saved_options.values() and v is not None
        }
        remote_params = raw_remote["parameters"]
        desired_params = [p.metadata() for p in self.parameters]
        effective_params = [
            (remote_params[i] | desired_params[i])
            if i < len(remote_params) and remote_params[i]["Name"] == desired_params[i]["Name"]
            else desired_params[i]
            for i, _ in enumerate(desired_params)
        ]
        remote_sql = textwrap.dedent(raw_remote["sql"].rstrip())
        remote_version = raw_remote["version"]
        desired_sql = textwrap.dedent(self.sql.rstrip())
        if (
            desired_sql == remote_sql
            and remote_props | desired_props == remote_props
            and effective_params == remote_params
        ):
            return None
        sql = self.template(desired)
        query(client, sql, self._backoff_handler)
        for i in range(1, self.Registration.tries + 1):
            version: int | None = self.registration_version(client, self.provider, self._backoff_handler)
            if version is not None and remote_version < version:
                break
            else:
                if i == self.Registration.tries:
                    sys.stderr.write(
                        f"warning: no view registration after {i} tries for {self.provider}"
                    )
                else:
                    time.sleep(self.Registration.wait_time)
        return {"provider": self.provider}

    def deps(self):
        return self.dependencies if self.dependencies else []

    @staticmethod
    def delete(client, old_state):
        sql = textwrap.dedent(f"""\
        @x = use Sys.Admin.SetupView
        --provider={old_state.provider}
        --deleteProvider
        ----
        select 1 as deleting
        enduse;
        select * from @x;
        """)
        backoff = BackoffHandler()
        query(client, textwrap.dedent(sql), backoff_handler=backoff)
        for i in range(1, ViewResource.Registration.tries + 1):
            if (
                ViewResource.registration_version(
                    client=client, view_name=old_state.provider, backoff_handler=backoff
                )
                is None
            ):
                break
            else:
                if i == ViewResource.Registration.tries:
                    sys.stderr.write(
                        f"warning: no view deregistration after {i} tries for {old_state.provider}"
                    )
                else:
                    time.sleep(ViewResource.Registration.wait_time)


def ser_key_key(value, info):
    if info.context and info.context.get("style") == "dump":
        return {"$ref": value.id}
    return value.provider


def des_key_key(value, info):
    if not isinstance(value, dict):
        return value
    if info.context and info.context.get("$refs"):
        ref = info.context["$refs"][value["$ref"]]
        return ref
    return value


ViewKey = Annotated[
    ViewResource | ViewRef,
    BeforeValidator(des_key_key), PlainSerializer(ser_key_key)
]


class InlineDataType(StrEnum):
    Text = "Text"
    Decimal = "Decimal"
    DateTime = "DateTime"
    Date = "Date"
    Boolean = "Boolean"
    Identifier = "_identifier"


class InlineProperty(CamelAlias, BaseModel):
    """A property that can be set inline in a view """
    key: PropertyKey | None = None  # set for properties not custom entities
    field: str | None = None        # set for structured properties and CEs, not scalars
    name: str
    description: str = ""
    data_type: InlineDataType | None = None
    is_main: None | bool = None
    as_at: None | datetime = None

    @field_serializer("as_at", when_used="always")
    def as_at_serializer(self, value: datetime | None) -> str | None:
        """Serialize the as_at datetime to a string in ISO 8601 format"""
        if value is None:
            return None
        return value.strftime("%Y-%m-%dT%H:%M:%S.%fZ")

    def prop_key(self):
        # key is set for a scalar property, key and field for a structured property
        # only field if we inlining a custom entity
        if self.key is not None and self.field is not None:
            return "/".join([self.key.domain, self.key.scope, self.key.code]) + "." + self.field
        elif self.key is not None:
            return "/".join([self.key.domain, self.key.scope, self.key.code])
        else:
            return self.field

    def sql_values(self):
        # key is set for a scalar property, key and field for a structured property
        # only field if we inlining a custom entity
        prop_key = self.prop_key()
        return [
            lumi_sql(prop_key),
            lumi_sql(self.name),
            lumi_sql(self.data_type),
            lumi_sql(self.description),
            lumi_sql(self.is_main),
            lumi_sql(self.as_at),
        ]


@register_resource()
class InlinePropertiesResource(BaseModel, Resource):
    id: str = Field(exclude=True)
    provider: str
    provider_name_extension: str | None = None
    properties: Sequence[InlineProperty]

    @model_validator(mode="before")
    @classmethod
    def des_model(cls, data: Any, info) -> Dict[str, Any]:
        if not isinstance(data, dict):
            return data
        if info.context and info.context.get("id"):
            return data | {"id": info.context.get("id")}
        return data

    @model_validator(mode="after")
    def validate_no_ext_custom(self):
        parts = self.provider.split(".")
        if len(parts) == 3 and parts[1].lower() == "customentity":
            if self.provider_name_extension is not None:
                raise KeyError("Cannot set use a provider name extension when creating "
                "a custom entity provider")
        return self

    @staticmethod
    def _convert_provider_name_to_path(provider, provider_name_extension) -> str:
        """Convert a provider name to a path format used in sys.file"""
        str_builder = ""
        if provider_name_extension:
            str_builder += f"{provider_name_extension.lower()}_"
        parts = provider.split(".")
        if len(parts) == 3 and parts[1].lower() == "customentity":
            str_builder += (
                parts[2] + "customentity"
            )
        else:
            str_builder += (
                provider.replace(".", "").lower().replace("lusid", "").replace("txn", "transaction")
            )
        return str_builder

    def _check_extension(self) -> str:
        """Check if the provider name extension is set and return the appropriate SQL clause"""
        if self.provider_name_extension:
            return f"and ProviderNameExtension = '{self.provider_name_extension}'"
        return ""

    def read(self, client: httpxClient, old_state: SimpleNamespace) -> None | Dict[str, Any]:
        # create backoff handler
        backoff = BackoffHandler()
        # run query to get the inline properties
        path = self._convert_provider_name_to_path(old_state.provider, old_state.extension)
        sql = textwrap.dedent(f"""\
            select Content from Sys.File
                where Path = 'config/lusid/factories/{path}providerfactory.csv'
            ;
        """)
        res = query(client, sql, backoff_handler=backoff)
        # If no results found, return None (properties don't exist yet)
        if not res:
            return None
        remote = self.parse_content(res)
        return remote

    def generate_sql(self):
        if len(self.properties) == 0:
            return textwrap.dedent(f"""\
                select * from Sys.Admin.Lusid.Provider.Configure
                    where Provider = '{self.provider}'
                    and WriteAction = 'Set'
                    {self._check_extension()}
                ;
            """)
        # build the values sql from the properties
        keys_right = ["(" + ", ".join(prop.sql_values()) + ")" for prop in self.properties]
        values_str = ",".join(keys_right)
        # build the columns clause
        columns_str = "column1 as [Key], column2 as Name, column3 as DataType, column4 as Description, "\
            "column5 as IsMain, column6 as AsAt"
        # set the inline properties of the provider to these values overwriting any existing ones
        return textwrap.dedent(f"""\
            @keysToCatalog = values
                {values_str}
            ;
            @config = select
                {columns_str}
                from @keysToCatalog
            ;
            select * from Sys.Admin.Lusid.Provider.Configure
                where Provider = '{self.provider}'
                and Configuration = @config
                and WriteAction = 'Set'
                {self._check_extension()}
            ;
        """)

    def parse_content(self, rows):
        csv_str = rows[0]["Content"]
        reader = csv.DictReader(StringIO(csv_str))
        data = [row for row in reader]
        # convert to a dictionary, sorted by key so we get consistent hashing
        return {
            prop["Key"]: {
                "Name": prop["Name"],
                "DataType": prop["DataType"],
                "Description": prop["Description"],
                "IsMain": prop.get("IsMain", None),
                "AsAt": prop.get("AsAt", None),
            }
            for prop in sorted(data, key=lambda x: x["Name"])
        }

    def create(self, client: httpxClient) -> None | Dict[str, Any]:
        # lumi query to create or update the provider columns
        backoff = BackoffHandler()
        sql = self.generate_sql()
        rows = query(client, sql, backoff_handler=backoff)
        if len(rows) > 0 and rows[0]["Error"]:
            raise RuntimeError("Failed to inline:" + rows[0]["Error"] + "\n" + sql)
        # parse the result
        properties_dict = self.parse_content(rows)
        remote_version = hashlib.sha256(json.dumps(properties_dict, sort_keys=True).encode()).hexdigest()
        # desired hash
        desired = self.model_dump(mode="json", by_alias=True, exclude_none=True)
        source_version = hashlib.sha256(json.dumps(desired, sort_keys=True).encode()).hexdigest()
        return {
            "provider": self.provider,
            "extension": self.provider_name_extension,
            "source_version": source_version,
            "remote_version": remote_version,
        }

    def update(self, client: httpxClient, old_state) -> Union[None, Dict[str, Any]]:
        # if the identity has changed, delete and recreate
        if self.provider != old_state.provider or self.provider_name_extension != old_state.extension:
            self.delete(client, old_state)
            return self.create(client)
        # check for changes
        remote = self.read(client, old_state)
        assert remote is not None
        remote_version = hashlib.sha256(json.dumps(remote, sort_keys=True).encode()).hexdigest()
        desired = self.model_dump(mode="json", by_alias=True, exclude_none=True)
        source_version = hashlib.sha256(json.dumps(desired, sort_keys=True).encode()).hexdigest()
        if source_version == old_state.source_version and remote_version == old_state.remote_version:
            return None
        return self.create(client)

    @staticmethod
    def delete(client: httpxClient, old_state) -> None:
        pne = f"and ProviderNameExtension = '{old_state.extension}'" if old_state.extension else ""
        backoff = BackoffHandler()
        sql = textwrap.dedent(f"""\
            select *
                from Sys.Admin.Lusid.Provider.Configure
                where Provider = '{old_state.provider}'
                and WriteAction = 'Reset'
                {pne}
            ;
        """)
        query(client, sql, backoff_handler=backoff)

    def deps(self) -> Sequence["Resource|Ref"]:
        deps = []
        for v in self.properties:
            if v.key:  # custom entity has no key
                deps.append(v.key)
        return deps



import os

import pytest

import fbnconfig
import tests.integration.identity as identity

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}


def test_teardown():
    deployment_name = identity.configure({}).id
    fbnconfig.deploy(fbnconfig.Deployment(deployment_name, []), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    matches = get_roles_by_name(client, "robTest_role2")
    assert len(matches) == 0
    users = get_users_by_login(client, "robtest_jane3@robtest.com")
    assert len(users) == 0


@pytest.mark.skip(reason="This creates actual users; should be run explicitly")
def test_create():
    fbnconfig.deploy(identity.configure(host_vars), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    matches = get_roles_by_name(client, "robTest_role2")
    assert len(matches) == 1
    users = get_users_by_login(client, "robtest_jane3@robtest.com")
    assert len(users) == 1


def test_update():
    fbnconfig.deploy(identity.configure(host_vars), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    matches = get_roles_by_name(client, "robTest_role2")
    assert len(matches) == 1
    users = get_users_by_login(client, "robtest_jane3@robtest.com")
    assert len(users) == 1


def get_roles_by_name(client, name):
    get = client.request("get", "/identity/api/roles")
    get.raise_for_status()
    roles = get.json()
    return [role for role in roles if role["name"] == name]


def get_users_by_login(client, login):
    get = client.request("get", "/identity/api/users")
    get.raise_for_status()
    users = get.json()
    return [user for user in users if user["login"] == login]



from fbnconfig import Deployment, access, identity, workspace


def configure(_):
    start_date = "2024-01-01T00:00:00.0000000+00:00"
    #
    # UK Trading Shared workspace
    #
    uk_trading_wksp = workspace.WorkspaceResource(
        id="workspace",
        visibility=workspace.Visibility.SHARED,
        name="example_trading_uk",
        description="workspace for UK Trading dashboards",
    )
    #
    # workspace Administrator
    #
    # this selector grants full control of the workspaces because action Any
    # includes ReadItem and WriteItem
    admin_selectors = [
        access.IdSelector(
            name="data_any_workspace_example_trading_uk",
            identifier={"visibility": "shared", "name": "example_trading_uk"},
            actions=[
                access.ActionId(scope="default", activity="Any", entity="Workspace"),
            ],
        )
    ]
    # policy to allow admin
    admin_data_policy = access.PolicyResource(
        id="admin_policy",
        code="data_admin_example_trading_uk",
        applications=["LUSID"],
        grant=access.Grant.ALLOW,
        selectors=admin_selectors,
        when=access.WhenSpec(activate=start_date),
        description="Allow full admin access to the UK trading workspace and items within",
    )
    # access role for admin to hold the permissions
    admin_access_role = access.RoleResource(
        id="admin_access_role",
        code="data_admin_example_trading_uk",
        resource=access.RoleResourceRequest(policy_id_role_resource=access.PolicyIdRoleResource(policies=[admin_data_policy])),
        when=access.WhenSpec(activate=start_date),
        permission=access.Permission.EXECUTE,
    )
    # identity role for admin to hold the users. Adding this role to a user will make them an admin
    admin_identity_role = identity.IdentityRoleResource(
        id="admin_identity_role",
        name="data_admin_example_trading_uk",
        description="irole workspace_data",
    )
    #
    # Workspace reader
    #
    # this selector grants read access to the workspace and read to the items
    # inside
    read_selectors = [
        access.IdSelector(
            name="data_read_workspace_example_trading_uk",
            identifier={"visibility": "shared", "name": "example_trading_uk"},
            actions=[
                access.ActionId(scope="default", activity="Read", entity="Workspace"),
                access.ActionId(scope="default", activity="ReadItem", entity="Workspace"),
            ],
        )
    ]
    # policy to allow read
    read_data_policy = access.PolicyResource(
        id="read_policy",
        code="data_read_example_trading_uk",
        applications=["LUSID"],
        grant=access.Grant.ALLOW,
        selectors=read_selectors,
        when=access.WhenSpec(activate=start_date),
        description="Allow read access to UK trading workspace and items",
    )
    # access role for read permissions
    read_access_role = access.RoleResource(
        id="read_access_role",
        code="data_read_example_trading_uk",
        resource=access.RoleResourceRequest(policy_id_role_resource=access.PolicyIdRoleResource(policies=[read_data_policy])),
        when=access.WhenSpec(activate=start_date),
        permission=access.Permission.EXECUTE,
    )
    # identity role for read. Adding this role to a user will allow them to read workspace items
    read_identity_role = identity.IdentityRoleResource(
        id="read_identity_role",
        name="data_read_example_trading_uk",
        description="Allow read access to UK trading workspace and items",
    )
    #
    # Deployment
    #
    return Deployment("workspace_example", [
        uk_trading_wksp,
        admin_data_policy, admin_access_role, admin_identity_role,
        read_data_policy, read_access_role, read_identity_role,

    ])



from fbnconfig import Deployment, side_definition, transaction_type

"""
An example configuration for defining transaction configuration related entities.
The script configures the following entities:
- Side
- Transaction type

More information can be found here:
https://support.lusid.com/knowledgebase/article/KA-01749/
"""


def configure(env):
    test_side1 = side_definition.SideResource(
        id="side1",
        side="Side1",
        scope="sc1",
        security="Txn:LusidInstrumentId",
        currency="Txn:SettlementCurrency",
        rate="Txn:TradeToPortfolioRate",
        units="Txn:Units",
        amount="Txn:TotalConsideration",
        notional_amount="0",
    )

    test_side2 = side_definition.SideResource(
        id="side2",
        side="Side2",
        scope="sc1",
        security="Txn:SettlementCurrency",
        currency="Txn:SettlementCurrency",
        rate="SettledToPortfolioRate",
        units="Txn:TotalConsideration",
        amount="Txn:TotalConsideration",
        notional_amount="0",
    )

    test_buy_transaction_type = transaction_type.TransactionTypeResource(
        id="example",
        scope="sc1",
        source="default",
        aliases=[
            transaction_type.TransactionTypeAlias(
                type="Buy",
                description="Example buy",
                transaction_class="default",
                transaction_roles="LongLonger",
                is_default=False,
            ),
            transaction_type.TransactionTypeAlias(
                type="BY",
                description="Example buy alias",
                transaction_class="default",
                transaction_roles="LongLonger",
                is_default=False,
            ),
        ],
        movements=[
            transaction_type.TransactionTypeMovement(
                movement_types=transaction_type.MovementType.StockMovement,
                side=test_side1,
                direction=1,
                name="Stock Movement",
                movement_options=[transaction_type.MovementOption.DirectAdjustment],
            )
        ],
        calculations=[
            transaction_type.TransactionTypeCalculation(
                type=transaction_type.CalculationType.TaxAmounts, side=test_side2
            )
        ],
    )

    return Deployment("transaction_configuration_example", [test_buy_transaction_type])
